/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#ifndef qKernel_H
#define qKernel_H
#if !(defined __XC16__) && !(defined __C30__)
    #error "qKernel for PIC24 is designed for the XC16 or C30 compiler"
#endif
#ifdef __XC16__
    #include <xc.h>
#endif
#ifdef __C30__
    #if defined(__dsPIC30F__)
            #include <p30Fxxxx.h>
    #elif defined(__dsPIC33E__)
            #include <p33Exxxx.h>
    #elif defined(__dsPIC33F__)
            #include <p33Fxxxx.h>
    #elif defined(__PIC24E__)
            #include <p24Exxxx.h>
    #elif defined(__PIC24F__)
            #include <p24Fxxxx.h>
    #elif defined(__PIC24FK__)
            #include <p24Fxxxx.h>
    #elif defined(__PIC24H__)
            #include <p24Hxxxx.h>
    #else
            #error 'Device not supported by qKernelV6.0'
    #endif
#endif
#include <stdint.h>
#include <stdbool.h>

#ifndef NULL
    #define NULL (0)
#endif

#define qBREAKPOINT                 __asm__ __volatile__ (" .pword 0xDA4000")
#define qBIT_SIZE                   16
#define qSTACK_GROWS_LOW_TO_HIGH 	1
#define qLITTLE_ENDIAN              1

#define TCB_STACK_GUARD_SIZE 		16
#define qNEAR                       __attribute__((near))


//****************************************************************************************************************
//*****************************************        Structures       **********************************************
//****************************************************************************************************************
struct sTCB {                       // Block is 96 bytes including context.
    struct sTCB* pNext;             // 000 Pointer to next TCB in the wait/ready list
    struct sTCB* pTcbList;          // 002 List of all TCB's. Starts with the idle thread
    char* pName;                    // 004 The name of the thread
    uint8_t Type;                   // 006 The type to check if this is a correct TCB (qTYPE_TCB)
    uint8_t ListType;               // 007 0=NoList, 1=Ready, 2=Timer, 3=Rtcc
    uint8_t TimedOut;               // 008 0=NOT-TimedOut 1=TimedOut
    uint8_t Prio;                   // 009 Priority (from 0 to 255)
    struct sTCB* pPrev;             // 010 Pointer to previous TCB in the wait/ready list
    uint32_t WakeUpTime;            // 012 Alarm time in cycles
    uint8_t Status;                 // 016 Status for the thread (xTCB_*)
    uint8_t TrackBit;               // 017 Bit to set for Tracking
    unsigned TBLPAG;                // 018 TBLPAG
    unsigned Tag;                   // 020 Tag free for use by the (advanced) developer
    struct sTCB *pNextWaitObj;      // 022 Next thread waits for object (Prio order) pObj must !=0
    struct sOBJ *pObject;           // 024 Object that we wait on (If !=0 we wait)
    unsigned* pStack;               // 026 Pointer to start of stack or offset in EDS pointer
    unsigned StackSize;             // 028 The size of the stack in bytes
    unsigned DSRPSVPAG;             // 030 DSRPAG or PSVPAG
    unsigned DSWPAG;                // 032 DSWPAG
    unsigned RCOUNT;                // 034 RCOUNT
    unsigned Registers[16];         // 036 Registers W0 to W15
    unsigned SPLIM;                 // 068 SPLIM
    unsigned WaitEvtEvents;         // 070 Events to wait for and return flags that were set
    unsigned WaitThrEvents;         // 072 Thread events to wait for and return flags that were set
    unsigned CurThrEvents;          // 074 Current Events state
    unsigned StackEdsPage;          // 076 Indicator and page for the EdsStack
    uint8_t* TrackAdr;              // 078 Filler
    uint64_t Cycles;                // 080 Total number of cycles by this process
    uint64_t LastCycles;            // 088 Cycles when becoming run-able (Must follow Cycles)
};                                  // 096 TOATAL SIZE
typedef struct sTCB * pTCB;

typedef struct sOBJ {               // Generic QKernel object EVT, QUE, MTX, PIP, SEM TMR
    struct sTCB *pWaitList;         // 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 
    uint8_t Filler007;              // 007 Filler of 1 byte
    void* Filler1;                  // 008 
} OBJ;                              // 010 Total Size
typedef OBJ* pOBJ;

typedef struct sEVT {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 qTYPE_EVT
    uint8_t Filler007;              // 007 Filler of 1 byte
    unsigned CurEvents;             // 008 The current state of events
} EVT;                              // 010 Total Size
typedef EVT* pEVT;

typedef struct sMTX {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 qTYPE_MTX
    uint8_t OrgPrio;                // 007 Original priority of the priority boosted thread
    pTCB Owner;                     // 008 The owner of the Mutex
} MTX;                              // 010 Total Size
typedef MTX* pMTX;

typedef struct sSEM {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 qTYPE_SEM
    uint8_t Filler007;              // 007 Filler of 1 byte
    int volatile Permits;           // 008 Available permits
} SEM;                              // 010 Total Size
typedef SEM* pSEM;

typedef struct sPUB {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 qTYPE_PUB
    uint8_t Filler007;              // 007 Filler of 1 byte
    struct sSUB *pSubscribers;      // 008 linked list of subscribers
} PUB;                              // 010 Total Size
typedef PUB* pPUB;

typedef struct sQUE {               // Variable Size object
    struct sTCB *pWaitList;         // 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 qTYPE_QUE
    uint8_t Status;                 // 007 Not used
    uint16_t Size;                  // 008 Size of the queue
    unsigned Entries;               // 010 Number of entries in the queue
    struct sMSG** pOut;             // 012 Index where to read the next message
    struct sMSG** pIn;              // 014 Index where to write the next message
    struct sMSG** pMax;             // 016 The end of the queue
    struct sMSG* pMessages;         // 018 The messages
} QUE;
typedef QUE* pQUE;

typedef struct sPIP{                // Variable Size object
    unsigned volatile Entries;      // 000 Number of entries in the pipe
    struct sOBJ *pObjList;          // 002 Next object in the ObjectList
    char* pName;                    // 004 The name of the object
    uint8_t Type;                   // 006 qTYPE_PIP
    uint8_t Status;                 // 007 Status NOT USED
    uint16_t BlockSizeINT;          // 008 Size of datatype/struct in integers (0 if not a multiply of integer)
    uint16_t BlockSize;             // 010 Size of the datatype or struct
    uint16_t MaxBlocks;             // 012 The max number of structures it holds - 1
    void (*pNtfWriter)(struct sPIP*, unsigned); // 014 Writing thread that can wait
    void (*pNtfReader)(struct sPIP*, unsigned); // 016 Reading thread that can wait
    uint8_t* pOut;                  // 018 Index where to read the next item
    uint8_t* pIn;                   // 020 Index where to write the next item
    uint8_t* pMax;                  // 022 The address paste the data
    uint8_t Data;                   // 024 Data
    uint8_t Filler025;              // 025 Filler of 1 byte
} PIP;                              // 026 Total Size
typedef PIP* pPIP;

typedef struct sTMR{                // Fixed Size object
    struct sTCB *pNext;             // 000 (Shared TCB) Pointer to next TMR/TCB in the wait/ready list
    struct sOBJ *pObjList;          // 002 (Shared OBJ) Next object in the ObjectList
    char* pName;                    // 004 (Shared OBJ) The name of the object
    uint8_t Type;                   // 006 (Shared TCB/OBJ) The type (qTYPE_TMR)
    uint8_t ListType;               // 007 (Shared TCB) 0=NoList, 1=Ready, 2=Timer, 3=Rtcc
    uint8_t TimedOut;               // 008 (Shared TCB) 0=NOT-TimedOut 1=TimedOut
    uint8_t Status;                 // 009 (Shared TCB) See xTMR_xxxxxxxx
    struct sTCB *Prev;              // 010 (Shared TCB) Pointer to previous TMR/TCB in the wait/ready list
    uint32_t WakeUpTime;            // 012 (Shared TCB) Alarm Time in cycles
    void (*pFunction)(void*);       // 016 The function to call
    int32_t WaitTime;               // 018 WaitTime during creation in cycles (Neg for RTCC waits)
    void* pParam;                   // 022 Parameter for function called when expired
} TMR;                              // 024 Total Size
typedef TMR* pTMR;

typedef struct sMBK {               // Size 2 or more (variable)
    struct sMPL *pMpl;              // 000 Pool Address to find the pool
    unsigned Memory;                // 002 Memory starts here
} MBK;
typedef MBK* pMBK;

typedef struct sMPL {               // Fixed Size object
    pMBK pFreeBlk;                  // 000 Pointer to the next free memory block
    struct sMPL *pNextPool;         // 002 Next memory pool
    uint8_t Type;                   // 004 The type (qTYPE_MPL)
    uint8_t Filler007;              // 005 Filler of 1 byte for 16 and 32 bit
    uint16_t BlockSize;             // 006 Size usable memory
    uint16_t Total;                 // 008 Total number of blocks currently in use
} MPL;                              // 010 Total Size
typedef MPL* pMPL;

typedef struct sFPL {               // Variable Size object (Data included)
    unsigned* pFirstBlk;            // 000 Pointer to the first memory block
    struct sFPL *pNextPool;         // 002 Next memory pool
    uint8_t Type;                   // 004 The type (qTYPE_FPL)
    uint8_t Filler007;              // 005 Filler of 1 byte
    uint16_t BlockSize;             // 006 Size usable memory
    uint16_t Total;                 // 008 Total number of blocks in use
    unsigned BitTable[];            // 010 at least one integer for bit-table 1=free 0=inuse
} FPL;
typedef FPL* pFPL;

typedef struct sSUB {               // Size 6 or 12
    struct sSUB *pNext;             // 000 The next subscriber
    void* pObject;                  // 002 Pipe or Queue
    void (*pFun)(void*, struct sMSG*);// 004 The subscriber function to call
} SUB;                              // 006 Total Size
typedef SUB* pSUB;

__extension__ typedef struct sMSG {
    union {
        struct sMSG* pNext;         // 000 Pointer to the next message
        struct sFPL* pFixedPool;    // 000 Pointer back to the fixed pool
    };
    uint8_t volatile UseCnt;        // 002 UseCount (READ-ONLY)
    uint8_t MemPoolType;            // 003 DO NOT USE
    uint16_t MsgType;               // 004 Type of message (READ-ONLY)
    unsigned Data[];                // 006 RO if UseCnt > 1 and RW if UseCnt == 1
} MSG;
typedef MSG* pMSG;

typedef struct sVER {               // Example 2.3-1100
    uint8_t major;                  // Major build number like 2
    uint8_t minor;                  // Minor build number like 3
    uint16_t build;                 // Build number 1100
} VER;
typedef const VER* pVER;

typedef struct sERR {
    uint32_t TrapAddress;           // 000 PC when trap occurred
    uint32_t Uptime;                // 004 Uptime in seconds
    pTCB pThread;                   // 008 The current thread
    uint16_t ErrNbr;                // 010 Specifies the last error
    uint8_t StackType;              // 012 Stack Type ThreadStack=0 InterruptStack=1
    uint8_t Zone;                   // 013 0=Thread, 1=Fiber, >1=Interrupt
    unsigned SPLIM;                 // 014 LIMIT of SP
    int InCritical;                 // 016 -1=no critical, 0=first critical etc.
} ERR;                              // 018 total length
typedef ERR* pERR;

typedef struct sBCD_DTM{
    uint8_t year;                   // BCD codification for year, 0x00 to 0x67
    uint8_t reserved;               // reserved for future use
    uint8_t mday;                   // BCD codification for day of month 0x01 to 0x31
    uint8_t mon;                    // BCD codification for month 0x01 to 0x31
    uint8_t hour;                   // BCD codification for hours 0x00 to 0x23
    uint8_t wday;                   // BCD codification for day of week 0x00 to 0x06
    uint8_t sec;                    // BCD codification for seconds 0x00 to 0x59
    uint8_t min;                    // BCD codification for minutes 0x00 to 0x59
} BCD_DTM;
typedef BCD_DTM* pBCD_DTM;

__extension__ typedef union {
    struct {
        uint8_t uint8_0;
        uint8_t uint8_1;
    };
    uint16_t uint16;
} uUINT16;
typedef uUINT16* pUINT16;

__extension__ typedef union {
    struct {
        uint8_t uint8_0;
        uint8_t uint8_1;
        uint8_t uint8_2;
        uint8_t uint8_3;
    };
    struct {
        uint16_t uint16_0;
        uint16_t uint16_1;
    };
    uint32_t uint32;
} uUINT32;
typedef uUINT32* pUINT32;

__extension__ typedef union {
    struct {
        uint8_t uint8_0;
        uint8_t uint8_1;
        uint8_t uint8_2;
        uint8_t uint8_3;
        uint8_t uint8_4;
        uint8_t uint8_5;
        uint8_t uint8_6;
        uint8_t uint8_7;
    };
    struct {
        uint16_t uint16_0;
        uint16_t uint16_1;
        uint16_t uint16_2;
        uint16_t uint16_3;
    };
    struct {
        uint32_t uint32_0;
        uint32_t uint32_1;
    };
    uint64_t uint64;                // 64 bits access
} uUINT64;
typedef uUINT64* pUINT64;

//*************************** All port functions *********************************
//*** Functions that are not in-line are in assembler or C modules ***************
//********************************************************************************
static inline __attribute__ ((always_inline)) unsigned qBit0TstSetAtomic(unsigned* Ptr) {
    unsigned ret=0;
    __asm__ __volatile__(
                "       btsts    [%1],#0            \n"
                "       bra      z, 1f              \n"
                "       inc      %0, %0             \n"
                "1:                                 \n"
                : "=r"(ret): "r"(Ptr) : "cc");
    return ret;
}
// The Bit in the qBit???Mem functions set/clr/tst/tgl functions describes the
// bit-number for port information. To address bit 8 at address 0x02c0 use bit-number
// 0x2c08
static inline __attribute__ ((always_inline)) void qBitMemSetAtomic(unsigned Bit) {
    __asm__ __volatile__(
                "        lsr %0, #4, w0             \n"
                "        bsw.z [w0], %0             \n"
                :: "r"(Bit) : "w0", "cc");
}
static inline __attribute__ ((always_inline)) void qBitMemClrAtomic(unsigned Bit) {
    __asm__ __volatile__(
                "        lsr %0, #4, w0             \n"
                "        cp  w0, w0                 \n"
                "        bsw.z [w0], %0             \n"
                :: "r"(Bit) : "w0", "cc");
}
static inline __attribute__ ((always_inline)) bool qBitMemTst(unsigned Bit) {
    bool ret;
    __asm__ __volatile__(
                "        lsr %1, #4, w0             \n"
                "        clr %0                     \n"
                "        btst.z [w0], %1            \n"
                "        btss SR,#1                 \n"
                "        inc %0, %0                 \n"
                : "=r"(ret): "r"(Bit) : "w0", "cc");
    return ret;
}
static inline __attribute__ ((always_inline)) void qBitMemTglAtomic(unsigned Bit) {
__asm__ __volatile__(
                "       lsr %0, #4, w0              \n"
                "       mov #1, w1                  \n"
                "       and #15, %0                 \n"
                "       sl  w1, %0, w1              \n"
                "       xor w1, [w0], [w0]          \n"
                : : "r"(Bit) : "w0", "w1", "cc");
}
//////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qBytClr(void* p, unsigned Size) {
    __asm__ __volatile__("        dec %1,%1                     \n"
                         "        repeat %1                     \n"
                         "        clr.b [%0++]                  \n"
                         : "+r"(p), "+r"(Size));
}
static inline __attribute__ ((always_inline)) void qBytDecAtomic(volatile uint8_t* p) {
    __asm__ __volatile__("dec.b  [%0],[%0] " : : "r"(p) : "cc");
}
static inline __attribute__ ((always_inline)) void qBytIncAtomic(volatile uint8_t* p) {
    __asm__ __volatile__("inc.b  [%0],[%0] " : : "r"(p) : "cc");
}
static inline __attribute__ ((always_inline)) void qBytMov(void* From, void* To, unsigned Size) {
    __asm__ __volatile__("        dec %2,%2                     \n"
                         "        repeat %2                     \n"
                         "        mov.b [%0++], [%1++]          \n"
                         : "+r"(From), "+r"(To), "+r"(Size));
}
static inline __attribute__ ((always_inline)) void qBytSet(void* p, unsigned Size) {
    __asm__ __volatile__("        dec %1,%1                     \n"
                         "        repeat %1                     \n"
                         "        setm.b [%0++]                 \n"
                         : "+r"(p), "+r"(Size));
}
//////////////////////////////////////////////////////////////////////////////////
static  inline __attribute__ ((always_inline)) void qCrtEnter() {
    __asm__ __volatile__("bclr IEC1, #11");
    __asm__ __volatile__("inc _qvInCritical");
}
static inline __attribute__ ((always_inline)) unsigned qCrtEnterFirst() {
    __asm__ __volatile__(
            "        bclr IEC1, #11           \n"
            "        inc _qvInCritical        \n"
            : : : "cc");
    return SR&2;    // returns true if in FIRST Critical
}
static inline __attribute__ ((always_inline)) void qCrtExit() {
    __asm__ __volatile__("dec _qvInCritical");
    __asm__ __volatile__("bset IEC1, #11");
}
//////////////////////////////////////////////////////////////////////////////////
uint32_t qDivU3216(uint32_t dividend, uint16_t divisor);
uint32_t qDivU3216R(uint32_t dividend, uint16_t divisor, uint16_t* Remainder);
uint64_t qDivU6416(uint64_t dividend, uint16_t divisor);
uint64_t qDivU6416R(uint64_t dividend, uint16_t divisor, uint16_t* Remainder);
uint16_t qDivScaling(uint16_t v1, uint16_t v2, uint16_t v3); // return (v1*v2)/v3
//////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qFbrEnqueue0(void (*pFbr)()) {
    __asm__ __volatile__("9:     mov      _qvFbrPut, w2         \n"
                         "       mov      [w2++],w3             \n"
                         "       mov      w3, _qvFbrPut         \n"
                         "       btsts    [w2],#0               \n"
                         "       bra      nz, 9b                \n"
                         "       add      %0,#1,[w2]            \n"
                         "       bset     _qvSpawn, #8          \n"
                         : : "r"(pFbr) : "w2", "w3");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrEnqueue1(void (*pFbr)(void*), void* pPar1) {
    __asm__ __volatile__("9:     mov      _qvFbrPut, w2         \n"
                         "       mov      [w2++],w3             \n"
                         "       mov      w3, _qvFbrPut         \n"
                         "       btsts    [w2],#0               \n"
                         "       bra      nz, 9b                \n"
                         "       mov      %1,[w2+2]             \n"
                         "       add      %0,#1,[w2]            \n"
                         "       bset     _qvSpawn, #8          \n"
                         : : "r"(pFbr), "r"(pPar1) : "w2", "w3");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrEnqueue2(void (*pFbr)(void*,void*), void* pPar1, void* pPar2) {
    __asm__ __volatile__("9:     mov      _qvFbrPut, w2         \n"
                         "       mov      [w2++],w3             \n"
                         "       mov      w3, _qvFbrPut         \n"
                         "       btsts    [w2],#0               \n"
                         "       bra      nz, 9b                \n"
                         "       mov      %1,[w2+2]             \n"
                         "       mov      %2,[w2+4]             \n"
                         "       add      %0,#1,[w2]            \n"
                         "       bset     _qvSpawn, #8          \n"
                         : : "r"(pFbr), "r"(pPar1), "r"(pPar2) : "w2", "w3");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrSpawn1() {
    __asm__ __volatile__("bset _qvSpawn, #3");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrSpawn2() {
    __asm__ __volatile__("bset _qvSpawn, #2");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrSpawn3() {
    __asm__ __volatile__("bset _qvSpawn, #1");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrSpawn4() {
    __asm__ __volatile__("bset _qvSpawn, #0");
    __asm__ __volatile__("bset IFS1, #11");
}
static inline __attribute__ ((always_inline)) void qFbrSpawnRtcc() {
    __asm__ __volatile__("bset _qvSpawn, #7");
    __asm__ __volatile__("bset IFS1, #11");
}
//////////////////////////////////////////////////////////////////////////////////
extern int qvInCritical;
static inline __attribute__ ((always_inline)) unsigned qKrnInCritical() {
    return qvInCritical!=-1;
}
static inline __attribute__ ((always_inline)) unsigned qKrnInFiber() {
    return (SR & (0xe0))==0x20;    // returns true if interrupt level = 1
}
unsigned  qKrnInIntStack();
static inline __attribute__ ((always_inline)) unsigned qKrnInIsr() {
    return (SR&0b11000000);
}
////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qMemFreeFast(void* p) {
    qCrtEnter();
    __asm__ __volatile__(
                "        mov	[--%0], w7             \n"
                "        mov    [w7], [%0]             \n"
                "        mov    %0, [w7]               \n"
                :: "r"(p): "w7");
    qCrtExit();
}

////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qSemReleaseFast(struct sSEM* Sem) {
    __asm__ __volatile__(
                "        inc	[%0], [%0]             \n"
                "        bra    gt, 3f                 \n"
                "        sub    #8, %0                 \n"
                "        call   _xSemRelease           \n"
                "3:                                    \n"
                :: "r"(Sem): "w0", "w1", "w2", "w3", "w4", "w5", "w6", "w7");
}
static inline __attribute__ ((always_inline)) void qSemAcquireFast(struct sSEM* Sem) {
    __asm__ __volatile__(
                "        dec	[%0], [%0]             \n"
                "        bra    ge, 3f                 \n"
                "        inc	[%0], [%0]             \n"
                "        push   [%0]                   \n"
                "        sub    #8, %0                 \n"
                "        call 	_xThrSuspendOnObject   \n"
                "        pop    [%0]                   \n"
                "        dec	[%0], [%0]             \n"
                "3:                                    \n"
                :: "r"(Sem): "w0", "w1", "w2", "w3", "w4", "w5", "w6", "w7");
}
////////////////////////////////////////////////////////////////////////////////
void      qThrDelay(unsigned USec);
static inline __attribute__ ((always_inline)) void qThrSuspendV4() {
    __asm__ __volatile__(
                "        bclr   IEC1, #11              \n"
                "        inc	_qvInCritical          \n"
                "        mov    _qvCurTcb, w4          \n"
                "        mov    [w4], w5               \n"
                "        mov    [w4+10], w6            \n"
                "        mov    w6, [w5+10]            \n"
                "        mov    [w4+10], w6            \n"
                "        mov    w5, [w6]               \n"
                "        clr    w5                     \n"
                "        mov.b  w5, [w4+7]             \n"
                "        bra    _ySwitchThread         \n"
                ::: "w4", "w5", "w6");
}
////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qWrdDecAtomic(volatile int* p) {
    __asm__ __volatile__("dec  [%0],[%0]" : : "r"(p) : "cc");
}
// if > zero decrement and return 1
// if zero return 0
// This only works for values between 0 and 2^15
static inline __attribute__ ((always_inline)) bool qWrdDecIfNotZero(volatile int* p) {
    bool ret;
    __asm__ __volatile__("    mov    #1, %0          \n"
                  "    dec [%1],[%1]          \n"
                  "    bra nn, 9f             \n"
                  "    inc [%1],[%1]          \n"
                  "    clr %0                 \n"
                  "9:                         \n"
                  : "=r"(ret) : "r"(p) : "cc");
    return ret;
}
static inline __attribute__ ((always_inline)) void qWrdIncAtomic(volatile int* p) {
    __asm__ __volatile__("inc  [%0],[%0]" : : "r"(p) : "cc");
}
static inline __attribute__ ((always_inline)) void qWrdMov(unsigned* From, unsigned* To, unsigned Size) {
    __asm__ __volatile__("        lsr.w %2,%2                   \n"
                         "        dec %2,%2                     \n"
                         "        repeat %2                     \n"
                         "        mov.w [%0++], [%1++]          \n"
                         : "+r"(From), "+r"(To), "+r"(Size));
}
static inline __attribute__ ((always_inline)) void qWrdClr(unsigned* p, unsigned Size) {
    __asm__ __volatile__("        lsr.w %1,%1                   \n"
                         "        dec %1,%1                     \n"
                         "        repeat %1                     \n"
                         "        clr.w [%0++]                  \n"
                         : "+r"(p), "+r"(Size));
}
static inline __attribute__ ((always_inline)) void qWrdSet(unsigned* p, unsigned Size) {
    __asm__ __volatile__("        lsr.w %1,%1                   \n"
                         "        dec %1,%1                     \n"
                         "        repeat %1                     \n"
                         "        setm.w [%0++]                 \n"
                         : "+r"(p), "+r"(Size));
}
/////////////////////////////////////////////////////////////////////////////////
void yKernelInit();
static inline __attribute__ ((always_inline)) void yKrnSetInterrupt() {
    __asm__ __volatile__ ("bset IFS1, #11");
}

// Kernel port functions
void      yThrStkInit(void (*Thread)(void*), void* data, pTCB Tcb, unsigned* Stack, unsigned StackSize);
unsigned  yMemSetup(unsigned InterruptStackSize);
void      yKernelInit();
void      ySwitchThread();
void      yStart();
pERR      yErrorInit();
void      yErrorReset(unsigned Err);
void      yModeIdle();
void      yModeSleep();
unsigned  yMemSetup(unsigned InterruptStackSize);
void      yModeIdle();
void      yModeSleep();




// Timer port functions
uint8_t   yTimerInit();
void      yTimerSetAlarm(uint32_t Alarm);
uint32_t  yTimerGetTime();
uint32_t  yTimerGetTimeAddUSec(uint32_t USec);
void      yTimerStop();
unsigned  yTimerReshuffleNeeded();
void      yTimerReshuffleEnd();
uint64_t  yGetCycles();



// The RTCC prototypes
uint8_t   yRtccInit() __attribute__((far));
void      yRtccSetDatTim(uint32_t DatTim) __attribute__((far));
uint32_t  yRtccGetDatTimRounded() __attribute__((far));
uint32_t  yRtccGetDatTim() __attribute__((far));
void      yRtccSetAlarm(uint32_t seconds) __attribute__((far));
uint32_t  yRtccGetUptime() __attribute__((far));

// Specific EDS modules
pTCB      qThrCreateEds(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio, unsigned* __attribute__((space(eds))) Stack);
pTCB      qThrCreateSuspendEds(char* pName, void (*Thread)(void*), void* pParam, unsigned StackSize, uint8_t Prio, unsigned* __attribute__((space(eds))) Stack);

//*********************************** End of Port functions ***********************

#ifdef __HAS_EDS__
    #define qISR(name)                                  \
        void _qqqIsr##name() {                          \
            __asm__ __volatile__(                       \
            "    .global _" #name "             \n"     \
            "_" #name ":                        \n"     \
            "    push    w0                     \n"     \
            "    mov     w15, w0                \n"     \
            "    cp      _qvIntStackLimit       \n"     \
            "    bra     gtu, 9f                \n"     \
            "    mov     _qvIntStack, w15       \n"     \
            "9:  push    w0                     \n"     \
            "    push    SPLIM                  \n"     \
            "    mov     _qvIntStackLimit,w0    \n"     \
            "    mov     w0, SPLIM              \n"     \
            "    push    w1                     \n"     \
            "    push.d  w2                     \n"     \
            "    push.d  w4                     \n"     \
            "    push.d  w6                     \n"     \
            "    push    RCOUNT                 \n"     \
            "    push    TBLPAG                 \n"     \
            "    push    DSRPAG                 \n"     \
            "    push    DSWPAG                 \n"     \
            "    call    __qIsr" #name "        \n"     \
            "    pop     DSWPAG                 \n"     \
            "    pop     DSRPAG                 \n"     \
            "    pop     TBLPAG                 \n"     \
            "    pop     RCOUNT                 \n"     \
            "    pop.d   w6                     \n"     \
            "    pop.d   w4                     \n"     \
            "    pop.d   w2                     \n"     \
            "    pop     w1                     \n"     \
            "    pop     SPLIM                  \n"     \
            "    pop     w15                    \n"     \
            "    pop     w0                     \n"     \
            "    retfie                         \n");   \
        }                                               \
        void _qIsr##name()                              \

    #define qISR_FAST(name)                             \
        void _qqqIsr##name() {                          \
            __asm__ __volatile__(                       \
            "    .global _" #name "             \n"     \
            "_" #name ":                        \n"     \
            "    push.s                         \n"     \
            "    mov    w15, w0                 \n"     \
            "    cp     _qvIntStackLimit        \n"     \
            "    bra    gtu, 9f                 \n"     \
            "    mov    _qvIntStack, w15        \n"     \
            "9:  push   w0                      \n"     \
            "    push   SPLIM                   \n"     \
            "    mov    _qvIntStackLimit,w0     \n"     \
            "    mov    w0, SPLIM               \n"     \
            "    push.d w4                      \n"     \
            "    push.d w6                      \n"     \
            "    push   RCOUNT                  \n"     \
            "    push   TBLPAG                  \n"     \
            "    push   DSRPAG                  \n"     \
            "    push   DSWPAG                  \n"     \
            "    call   __qIsr" #name "         \n"     \
            "    pop    DSWPAG                  \n"     \
            "    pop    DSRPAG                  \n"     \
            "    pop    TBLPAG                  \n"     \
            "    pop    RCOUNT                  \n"     \
            "    pop.d  w6                      \n"     \
            "    pop.d  w4                      \n"     \
            "    pop    SPLIM                   \n"     \
            "    pop    w15                     \n"     \
            "    pop.s                          \n"     \
            "    retfie                         \n");   \
        }                                               \
        void _qIsr##name()                              \

#else
    #define qISR(name)                                  \
        void _qqqIsr##name() {                          \
            __asm__ __volatile__(                       \
            "    .global _" #name "             \n"     \
            "_" #name ":                        \n"     \
            "    push    w0                     \n"     \
            "    mov     w15, w0                \n"     \
            "    cp      _qvIntStackLimit       \n"     \
            "    bra     gtu, 9f                \n"     \
            "    mov     _qvIntStack, w15       \n"     \
            "9:  push    w0                     \n"     \
            "    push    SPLIM                  \n"     \
            "    mov     _qvIntStackLimit,w0    \n"     \
            "    mov     w0, SPLIM              \n"     \
            "    push    w1                     \n"     \
            "    push.d  w2                     \n"     \
            "    push.d  w4                     \n"     \
            "    push.d  w6                     \n"     \
            "    push    RCOUNT                 \n"     \
            "    push    TBLPAG                 \n"     \
            "    push    PSVPAG                 \n"     \
            "    call    __qIsr" #name "        \n"     \
            "    pop     PSVPAG                 \n"     \
            "    pop     TBLPAG                 \n"     \
            "    pop     RCOUNT                 \n"     \
            "    pop.d   w6                     \n"     \
            "    pop.d   w4                     \n"     \
            "    pop.d   w2                     \n"     \
            "    pop     w1                     \n"     \
            "    pop     SPLIM                  \n"     \
            "    pop     w15                    \n"     \
            "    pop     w0                     \n"     \
            "    retfie                         \n");   \
        }                                               \
        void _qIsr##name()
    #define qISR_FAST(name)                             \
        void _qqqIsr##name() {                          \
            __asm__ __volatile__(                       \
            "    .global _" #name "             \n"     \
            "_" #name ":                        \n"     \
            "    push.s                         \n"     \
            "    mov     w15, w0                \n"     \
            "    cp      _qvIntStackLimit       \n"     \
            "    bra     gtu, 9f                \n"     \
            "    mov     _qvIntStack, w15       \n"     \
            "9:  push    w0                     \n"     \
            "    push    SPLIM                  \n"     \
            "    mov     _qvIntStackLimit,w0    \n"     \
            "    mov     w0, SPLIM              \n"     \
            "    push.d  w4                     \n"     \
            "    push.d  w6                     \n"     \
            "    push    RCOUNT                 \n"     \
            "    push    TBLPAG                 \n"     \
            "    push    PSVPAG                 \n"     \
            "    call    __qIsr" #name "        \n"     \
            "    pop     PSVPAG                 \n"     \
            "    pop     TBLPAG                 \n"     \
            "    pop     RCOUNT                 \n"     \
            "    pop.d   w6                     \n"     \
            "    pop.d   w4                     \n"     \
            "    pop     SPLIM                  \n"     \
            "    pop     w15                    \n"     \
            "    pop.s                          \n"     \
            "    retfie                         \n");   \
        }                                               \
        void _qIsr##name()
#endif

#if defined(__HAS_EDS__)
    extern unsigned* qvSharedStack;     // Working stack for all EDS stacks
    extern pTCB qvSharedStackOwner;     // The owner of the shared stack
    extern unsigned qvSharedStackSize;  // The size of the shared stack
#endif



#include "../qKernelCommon.h"


#endif //qKernel_H


