/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  
#include "yKernel.h"
//     
// This module handles the RTCC with the 0.5 second tick provided by TIMER1. 
//

uint32_t qvUpTime = 0;            		// UpTime of the device
uint32_t qvAlarm = 0;             		// The alarm Date-Time in DTM format (nbr sec since 1/1/2000 00:00:00)
uint32_t qvDatTim = 0;            		// Current DateTime in emulation mode
uint8_t  qvHalfSec = 0;           		// 0=in first half  1=second half

uint8_t yRtccInit() {
    __builtin_write_OSCCONL(OSCCON | 0x2);

    T1CON = 0;
    TMR1 = 0;
    PR1 = 0x4000;                       // 0x4000=16384  32768/16384 = 2Hz 
    IPC0bits.T1IP = 2;                  // Interrupt priority 2
    IFS0bits.T1IF = 0;                  // Clear the flag
    IEC0bits.T1IE = 1;                  // Set the interrupt
    T1CON = 0x8002;                     // Timer ON and External oscillator
    return 2;
}    

//
// This function sets the current time
//
void yRtccSetDatTim(uint32_t DatTim) {
    qCrtEnter();
    qvDatTime = DatTim;
    qCrtExit();
}

//
// This function returns the current time to the user
//
uint32_t yRtccGetDatTim() {
    uint32_t datTim;
    qCrtEnter();
    datTim = qvDatTime;
    qCrtExit();
    return datTim;
}    

//
// This function returns the current time to the user + 1 if in the second part of the half Second
//
uint32_t yRtccGetDatTimRounded() {
    uint32_t datTim;
    qCrtEnter();
    datTim = qvDatTime + qvHalfSec;
    qCrtExit();
    return datTim;
}    

//
// This function set the alarm time
//
void yRtccSetAlarm(uint32_t alarm){
    qCrtEnter();
    qvAlarm = alarm;
    qCrtExit();
}    

//
// This function returns the number of seconds the system is up.
//
uint32_t yRtccGetUptime() {
    uint32_t upTime;
    qCrtEnter();
    upTime = qvUpTime;
    qCrtExit();
    return upTime;
}    

//
// Interrupt by the Timer1 every 0.5 second.
//
void __attribute__((__interrupt__, no_auto_psv)) _T1Interrupt(void) {
    IFS0bits.T1IF = 0;
    qCrtEnter();
    if (qvHalfSec==0) {
        qvHalfSec=1;
        qCrtExit();
        return;
    }        
    qvHalfSec=0;
    qvDatTime++;
    qvUpTime++;
    if (qvAlarm!=0) {
        if (qvDatTime>=qvAlarm) {
            qvAlarm = 0;
            qFbrSpawnRtcc();
        }
    }
    qCrtExit();
}    

