/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "yKernel.h"

// This example implements the RTCC in hardware. The RTCC variables are not
// updated.

#define xRtccOn()                            (RCFGCALbits.RTCEN=1)
#define xRtccAlrmEnable()                    (ALCFGRPTbits.ALRMEN=1)
#define xRtccAlrmDisable()                   (ALCFGRPTbits.ALRMEN=0)
#define xRtccIs2ndHalfSecond()               (RCFGCALbits.HALFSEC)
#define xRtccClearRtcPtr()                   (RCFGCAL&=~RCFGCAL_MASK_RTCPTR)
#define xRtccSetRtcPtr(mask)                 (RCFGCAL|=mask)
#define xRtccSetAlrmPtr(mask)                (ALCFGRPT |= mask)
#define xRtccClearAlrmPtr()                  (ALCFGRPT&=~ALCFGRPT_MASK_ALRMPTR)
#define xRtccSetInt(enable)                  (IEC3bits.RTCIE=enable)
#define xRtccSetIntPriority(pri)             (IPC15bits.RTCIP=pri)
#define xRtccClearIntrStatusBit()            (IFS3bits.RTCIF = 0)



// A MCU is defined so we have all the definitions
typedef enum
{
    RCFGCAL_MASK_RTCEN      =   0x8000, 
    RCFGCAL_MASK_FRZ        =   0x4000,
    RCFGCAL_MASK_RTCWREN    =   0x2000,
    RCFGCAL_MASK_RTCSYNC    =   0x1000,
    RCFGCAL_MASK_HALFSEC    =   0x0800,
    RCFGCAL_MASK_RTCOE      =   0x0400,
    RCFGCAL_MASK_RTCPTR     =   0x0300,
    RCFGCAL_MASK_CAL        =   0x00ff
}RCFGCAL_MASK;

//Alarm definitions
typedef enum
{
    ALCFGRPT_MASK_ALRMEN    =   0x8000,
    ALCFGRPT_MASK_CHIME     =   0x4000,
    ALCFGRPT_MASK_AMASK     =   0x3c00,
    ALCFGRPT_MASK_ALRMPTR   =   0x0300,
    ALCFGRPT_MASK_ARPT      =   0x00ff
}ALCFGRPT_MASK;

// accessing the RTCC/Alarm Value Register Window pointer bits
typedef enum
{
    RTCCPTR_MASK_SECMIN     =   0x0000,
    RTCCPTR_MASK_HRSWEEK    =   0x0100,
    RTCCPTR_MASK_DAYMON     =   0x0200,
    RTCCPTR_MASK_YEAR       =   0x0300   
}RTCCPTR_MASK;



uint8_t yRtccInit() {
    qvSpawnRtccExpire = xFbrRTCCExpire;         // This function is called by the Spawn

    // enable the Secondary Oscillator
    __builtin_write_OSCCONL(OSCCON | 0x2);
    RCFGCAL = 0x0;
    while (RCFGCALbits.RTCWREN==0) {            // Repeat in case it was interrupted
        asm volatile ("push w0");
        asm volatile ("mov #0x55, w0");
        asm volatile ("mov w0, _NVMKEY");       // Write the 0x55 key
        asm volatile ("mov #0xAA, w0");
        asm volatile ("mov w0, _NVMKEY");       // Write the 0xAA key
        asm volatile ("bset _RCFGCAL, #13");    // set the RTCWREN bit
        asm volatile ("pop w0");
    }
    xRtccSetIntPriority(2);
    yRtccSetDatTim(DTM_1_1_2010);               // Put a useful date in.
    return 3;
}    


void yRtccSetDatTim(uint32_t DatTim) {
    BCD_DTM dt;
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    unsigned* p = (unsigned*)&dt;

    qCrtEnter();
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    qDtmToBcdDtm(DatTim, &dt);
    xRtccClearRtcPtr();
    xRtccSetRtcPtr(RTCCPTR_MASK_YEAR);
    RTCVAL = *(p+0);
    RTCVAL = *(p+1);
    RTCVAL = *(p+2);
    RTCVAL = *(p+3);
    xRtccOn();
    xRtccSetInt(1);
    qCrtExit();
}

    
//
// This function returns the current time to the user
//
uint32_t yRtccGetDatTim() {
    BCD_DTM dt1;
    BCD_DTM dt2;
    // Ignore the warning "Taking the address of 'dt1' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    unsigned* p1 = (unsigned*)&dt1;
    // Ignore the warning "Taking the address of 'dt2' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    unsigned* p2 = (unsigned*)&dt2;

    do { // read until they produce the same value
        xRtccClearRtcPtr();
        xRtccSetRtcPtr(RTCCPTR_MASK_YEAR);
        *(p1+0) = RTCVAL;
        *(p1+1) = RTCVAL;
        *(p1+2) = RTCVAL;
        *(p1+3) = RTCVAL;
        // second read
        xRtccClearRtcPtr();
        xRtccSetRtcPtr(RTCCPTR_MASK_YEAR);
        *(p2+0) = RTCVAL;
        *(p2+1) = RTCVAL;
        *(p2+2) = RTCVAL;
        *(p2+3) = RTCVAL;
    } while (dt1.sec != dt2.sec) ;
    // Ignore the warning "Taking the address of 'dt1' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    return qDtmFromBcdDtm(&dt1);
}    

//
// This function returns the current time to the user + 1 if in the second part of the half Second
//
uint32_t yRtccGetDatTimRounded() {
    uint32_t datTim;
    qCrtEnter();
    datTim = yRtccGetDatTim();
    if (xRtccIs2ndHalfSecond()) datTim++;
    qCrtExit();
    return datTim;
}    

//
// This function set the alarm time
//
void yRtccSetAlarm(uint32_t alarm){
    BCD_DTM dt;
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    unsigned* p = (unsigned*)&dt;
    
    // Ignore the warning "Taking the address of 'dt' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    qDtmToBcdDtm(alarm, &dt);

    qCrtEnter();
    xRtccAlrmDisable();            // disable so no unexpected interrupts
    xRtccClearAlrmPtr();
    xRtccSetAlrmPtr(RTCCPTR_MASK_DAYMON);
    ALRMVAL = *(p+1);
    ALRMVAL = *(p+2);
    ALRMVAL = *(p+3);
    ALCFGRPTbits.AMASK = 9;

    xRtccClearIntrStatusBit();
    xRtccAlrmEnable();                // enable the alarm
    qCrtExit();
}    

//
//    This function returns the number of seconds the system is up.
//
uint32_t yRtccGetUptime() {
    return 0;
}    

//
// Interrupt by the RTCC. It will spawn a fiber to indicate that the time has been expired
//
void __attribute__ ((__interrupt__, no_auto_psv)) _RTCCInterrupt(void){
    xRtccClearIntrStatusBit();
    qFbrSpawnRtcc();
}




