/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

.macro SET_KERNEL_INTERRUPT
	bset IFS1, #11
.endm
.macro CLR_KERNEL_INTERRUPT
	bclr IFS1, #11
.endm
.macro TST_KERNEL_INTERRUPT
	btst IFS1, #11
.endm
.macro qCrtEnter
	bclr IEC1, #11
	inc		_qvInCritical
.endm
.macro qCrtExit
	dec		_qvInCritical
	bset IEC1, #11
.endm
.macro DISABLE_KERNEL_INTERRUPT
	bclr IEC1, #11
.endm
.macro ENABLE_KERNEL_INTERRUPT
	bset IEC1, #11
.endm

_xInitDone = 0
_xOsStarted = 1
_xStatisticsOn = 2
_xSwitchNotificationOn = 3
_xUSecOn = 4
_xPriorityCheckOn = 5
_xTimerInterruptOn = 6
_xRtccInterruptOn = 7


.section *,data,near  
_qvPR4:				.word 0
_qvPR5:				.word 0

.section qKernel,code

.global _yTimerInit
_yTimerInit:
	bclr	_T4CON, #15				;***** set timer off
	bset	_T4CON, #3				;***** 32 bit timer
	clr		_TMR5HLD				;*****
	clr		_TMR4					;*****
	setm	_PR4					;*****	
	setm	_PR5					;*****
	mov		#0xfff0, w0				;***** flags _
	and		_IPC7					;***** clear IP5 bits
	mov		#0x0002, w0				;***** interrupt priority 2 _
	ior		_IPC7					;***** in register
	bclr 	_IFS1, #12				;***** clear interrupt flag
	bset	_IEC1, #12				;***** Enable interrupt
    retlw	#4, w0                  ;***** We use TRM4/TMR5
	
	
.global __T5Interrupt				;*****
__T5Interrupt:						;*****
	push	w0
	mov		_PR5, w0				;*****
	setm	_PR5					;***** Prevents another interrupt
	mov		w0, _qvPR5
	mov		_PR4, w0				;*****
	mov		w0, _qvPR4
	add		_qvTotalTMR
	mov		_qvPR5, w0				;*****
	addc	_qvTotalTMR+2
	bclr 	_IFS1, #12				;*****
	bset 	_qvSpawn, #5			; set expire timer xFbrTimerExpire	
	bset	_qvKernel, #_xTimerInterruptOn
	SET_KERNEL_INTERRUPT			; Set the kernel interrupt (macro)
	mov		_qvPR4, w0				; add cycles
	add		_qvCycles+0
	mov		_qvPR5, w0
	addc	_qvCycles+2
	clr		w0
	addc	_qvCycles+4
	addc	_qvCycles+6
	pop		w0
	retfie

;
; void yTimerStop();
;
.global _yTimerStop
_yTimerStop:
	setm	_PR4					;***** Prevents interrupts
	bclr	_T4CON, #15				;***** set timer off
	clr		_TMR5HLD				;*****
	clr		_TMR4					;*****
	clr		_qvTotalTMR
	clr		_qvTotalTMR+2
	clr		_qvCycles
	clr		_qvCycles+2
	clr		_qvCycles+4
	clr		_qvCycles+6
	return

	
;
; unsigned yTimerReshuffleNeeded();
;
.global _yTimerReshuffleNeeded
_yTimerReshuffleNeeded:
	bclr	_IEC1, #12				;***** Disable interrupt
	btsc	_qvTotalTMR+2, #15		; More than 2^31
	retlw	#1, w0					;
	bset	_IEC1, #12				;***** Enable interrupt
	retlw	#0, w0
;
; void yTimerReshuffleEnd();
;
.global _yTimerReshuffleEnd
_yTimerReshuffleEnd:
	bclr	_qvTotalTMR+2, #15		;
	bset	_IEC1, #12				;***** Enable interrupt
	return
;
; uint32_t yTimerGetTime();
;
.global _yTimerGetTime
1:
	bset	_IEC1, #12				;***** Enable interrupt
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
_yTimerGetTime:
	bclr	_IEC1, #12				;***** Disable interrupt
	mov		_TMR4, w0				;*****
	mov		_TMR5HLD, w1			;*****
	mov		_qvTotalTMR, w2
	mov		_qvTotalTMR+2, w3
	add		w2, w0, w0
	addc	w3, w1, w1
;
	btst	_IFS1, #12				;***** Is the interrupt set
	bra		nz, 1b
	bset	_IEC1, #12				;***** Enable interrupt
	return
;
; uint32_t yTimerGetTimeAddUSec(uint32_t USec);
;
.global _yTimerGetTimeAddUSec
1:
	bset	_IEC1, #12				;***** Enable interrupt
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
_yTimerGetTimeAddUSec:
	bclr	_IEC1, #12				;***** Disable interrupt
	mov		_TMR4, w4				;*****
	mov		_TMR5HLD, w5			;*****
	mov		_qvTotalTMR, w2
	mov		_qvTotalTMR+2, w3
	add		w2, w4, w2
	addc	w3, w5, w3              ; result in W2-W3. USec still in w0-w1
;
	btst	_IFS1, #12				;***** Is the interrupt set
	bra		nz, 1b
	bset	_IEC1, #12				;***** Enable interrupt
    ; Make USec to cycles
    cp0     w1                      ; Is the number of uSec > 65k
    bra     nz, 5f
;
; Here we first Multiply and then divide
;
	mov		_qvClockFrequencyDiv15625, w6
	mul.uu	w1, w6, w4
	mul.uu	w0, w6, w0
	add		w4, w1, w1						; result in w0:w1
	; The result we divide by 64 (is >>6) to get
	; uSec*(Freq/15625/64) is uSec*(Freq/1000000) the number of cycles
	sl		w1, #10, w4
	lsr		w0, #6, w0
	ior		w4, w0, w0
	lsr		w1, #6, w1
    ; Add the GetTime (w2-w3) to the cycles
    add     w2, w0, w0
    addc    w3, w1, w1
    return
;
; Here we first divide by 64 and later multiply by freq/15625 to prevent overflow
;
5:	sl		w1, #10, w4             ; First shift 6 to the right
	lsr		w0, #6, w0              ; uSec/64
	ior		w4, w0, w0
	lsr		w1, #6, w1
	mov		_qvClockFrequencyDiv15625, w6
	mul.uu	w1, w6, w4              ; Now multiply ClockFreq/15625
	mul.uu	w0, w6, w0
	add		w4, w1, w1				; result in w0:w1 (uSec/64)*(Freq/15625)
                                    ; is uSec*Freq/1000000 is cycles
    ; Add the GetTime (w2-w3) to the cycles
    add     w2, w0, w0
    addc    w3, w1, w1
    return
;
; uint32_t yTimerGetTimeAddCycles(uint32_t USec);
;
.global _yTimerGetTimeAddCycles
1:
	bset	_IEC1, #12				;***** Enable interrupt
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
_yTimerGetTimeAddCycles:
	bclr	_IEC1, #12				;***** Disable interrupt
	mov		_TMR4, w4				;*****
	mov		_TMR5HLD, w5			;*****
	mov		_qvTotalTMR, w2
	mov		_qvTotalTMR+2, w3
	add		w2, w4, w2
	addc	w3, w5, w3              ; result in W2-W3. Cycles still in w0-w1
    ; Add the GetTime (w2-w3) to the cycles
    add     w2, w0, w0
    addc    w3, w1, w1
	return
;
; void yTimerSetAlarm(uint32_t TimeAlarm);
; Disable the interrupts and just subtract the new-time from qvTotalTMR
; Store in the PR value and enable interrupts.
; It is posible that TMR is to close to PR so we set the kernel interrupt 
; The function xFbrTimerExpire() will deal with it.
;
.global _yTimerSetAlarm
1:
	bset	_IEC1, #12				;***** Enable interrupt
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
_yTimerSetAlarm:
	bclr	_IEC1, #12				;***** Disable interrupt
	mov		_qvTotalTMR, w2
	mov		_qvTotalTMR+2, w3
	sub		w0, w2, w4				; w4/w5 = new time - qvTotalTMR
	subb	w1, w3, w5
	setm	_PR5					;***** This blocks the timer from readying
	mov		w4, _PR4				;*****
	mov		w5, _PR5				;*****
	mov		_TMR4, w6				;***** Get the TMR again
	mov		_TMR5HLD, w7			;*****
    add     #1000, w6               ; It is possible that we are hit by an interrupt
    addc    #0, w7                  ; We add 1000 cycles for that and that should be enough
    cp      w6, w4
    cpb     w7, w5
    bra     leu, 2f                 
	bset 	_qvSpawn, #5			; set expire timer xFbrTimerExpire	in case it is to close 
	SET_KERNEL_INTERRUPT			; Set the kernel interrupt (macro)
2:	btst	_IFS1, #12				;***** Is the interrupt set
	bra		nz, 1b
	bset	_IEC1, #12				;***** Enable interrupt
	bset	_T4CON, #15				;***** set timer on in case it was off
	return

;
;
; uint64_t yGetCycles()
;	Uses:	w0:w5
;
.global _yGetCycles
1:	bset	_IEC1, #12				;***** Enable interrupt
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
	nop								; This is to give the ISR time to start
; This is the entry point
_yGetCycles:
	bclr	_IEC1, #12				;***** Disable interrupt
	mov		_TMR4, w4				;***** timer value 
	mov		_TMR5HLD, w5			;***** timer value 
	mov		_qvCycles+0, w0			; Move to register
	mov		_qvCycles+2, w1			; Move to register
	mov		_qvCycles+4, w2			; Move to register
	mov		_qvCycles+6, w3			; Move to register
	add		w4, w0, w0
	addc	w5, w1, w1
	addc	#0, w2
	addc	#0, w3					; addded result in w0-w3 used for return variable
;	
	btst	_IFS1, #12				;***** Interrupt set?
	bra		nz, 1b
	bset	_IEC1, #12				;***** Enable interrupt
	return							; done w0-w3 contains cycles



