/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/

#ifndef qKernel_H
#define qKernel_H
#if !(defined __XC32__) && !(defined __C32__)
    #error "qKernel for PIC32 is designed for the XC32 or C32 compiler"
#endif
#ifdef __XC32__
    #include <xc.h>
#endif
#ifdef __C32__
    #include <p32xxxx.h>
#endif

#include <stdint.h>
#include <stdbool.h>

#ifndef NULL
    #define NULL (0)
#endif

#define qBREAKPOINT                     __asm__ __volatile__ (" sdbbp 0")
#define qBIT_SIZE                       32
#define qSTACK_GROWS_LOW_TO_HIGH        0
#define qLITTLE_ENDIAN                  1
#define qNEAR


//****************************************************************************************************************
//*****************************************        Structures       **********************************************
//****************************************************************************************************************
struct sTCB {                           // Block is 216 bytes including statistics
    struct sTCB* pNext;                 // 000 Pointer to next TCB in the wait/ready list
    struct sTCB* pTcbList;              // 004 List of all TCB's. Starts with the idle list
    char* pName;                        // 008 The name of the thread
    uint8_t Type;                       // 012 The type to check if this is a correct TCB (qTYPE_TCB)
    uint8_t ListType;                   // 013 0=NoList, 1=Ready, 2=Timer, 3=Rtcc
    uint8_t TimedOut;                   // 014 0=NOT-TimedOut 1=TimedOut
    uint8_t Prio;                       // 015 Priority (from 0 to 255)
    struct sTCB *pPrev;                 // 016 Pointer to previous TCB in the wait/ready list
    uint32_t WakeUpTime;                // 020 Alarm time
    uint8_t Status;                     // 024 Status for the thread (qqTCB_*)
    uint8_t TrackBit;                   // 025 Thread tracking bit number
    uint16_t Filler026;                 // 026 Filler
    uint8_t* TrackAdr;                  // 028 Track Address
    unsigned Tag;                       // 032 Tag free for use by the (advanced) developer
    struct sTCB *pNextWaitObj;          // 036 Next thr waits for object (Prio order) pObj must !=0
    struct sOBJ *pObject;               // 040 Object that we wait on (If !=0 we wait)
    unsigned* pStack;                   // 044 The start of the stack
    unsigned StackSize;                 // 048 The size of the stack in bytes
    unsigned at;                        // 052 
    unsigned v0;                        // 056 
    unsigned v1;                        // 060 
    unsigned a0;                        // 064 
    unsigned a1;                        // 068 
    unsigned a2;                        // 072 
    unsigned a3;                        // 076 
    unsigned t0;                        // 080 
    unsigned t1;                        // 084 
    unsigned t2;                        // 088 
    unsigned t3;                        // 092 
    unsigned t4;                        // 096 
    unsigned t5;                        // 100 
    unsigned t6;                        // 104 
    unsigned t7;                        // 108 
    unsigned t8;                        // 112
    unsigned t9;                        // 116
    unsigned s0;                        // 120
    unsigned s1;                        // 124
    unsigned s2;                        // 128
    unsigned s3;                        // 132
    unsigned s4;                        // 136 
    unsigned s5;                        // 140
    unsigned s6;                        // 144
    unsigned s7;                        // 148
    unsigned s8;                        // 152
    unsigned ra;                        // 156 
    unsigned HI;                        // 160 
    unsigned LO;                        // 164 
    unsigned SP;                        // 168 
    unsigned SR;                        // 172
    unsigned EPC;                       // 176
    unsigned WaitEvtEvents;             // 188 Events to wait for and return flags that were set
    unsigned WaitThrEvents;             // 192 Thread events to wait for and return flags that were set
    unsigned CurThrEvents;              // 196 Current Events state
    uint64_t Cycles;                    // 200
    uint64_t LastCycles;                // 208
};                                      // Total size 216 bytes
typedef struct sTCB * pTCB;

typedef struct sOBJ {               // Generic QKernel object EVT, QUE, MTX, PIP, SEM TMR
    struct sTCB *pWaitList;         // 000 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012
#if qBIT_SIZE==16
    uint8_t Filler007;              // 007 013 Filler of 1 byte
#else
    uint8_t Filler013[3];           // 007 013 Filler of 3 bytes
#endif
    void* Filler1;                  // 008 016
} OBJ;                              // 010 020 Total Size
typedef OBJ* pOBJ;

typedef struct sEVT {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012 qTYPE_EVT
#if qBIT_SIZE==16
    uint8_t Filler007;              // 007 013 Filler of 1 byte
#else
    uint8_t Filler013[3];           // 007 013 Filler of 3 bytes
#endif
    unsigned CurEvents;             // 008 016 The current state of events
} EVT;                              // 010 020 Total Size
typedef EVT* pEVT;

typedef struct sMTX {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012 qTYPE_MTX
    uint8_t OrgPrio;                // 007 013 Original prio of the priority boosted thread
    #if qBIT_SIZE==32
        uint16_t Filler014;         // 008 014 Filler of 2 bytes
    #endif
    pTCB Owner;                     // 008 016 The owner of the Mutex
} MTX;                              // 010 020 Total Size
typedef MTX* pMTX;

typedef struct sSEM {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012 qTYPE_SEM
#if qBIT_SIZE==16
    uint8_t Filler007;              // 007 013 Filler of 1 byte
#else
    uint8_t Filler013[3];           // 007 013 Filler of 3 bytes
#endif
    int volatile Permits;           // 008 016 Available permits
} SEM;                              // 010 020 Total Size
typedef SEM* pSEM;

typedef struct sPUB {               // Fixed Size object
    struct sTCB *pWaitList;         // 000 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012 qTYPE_PUB
#if qBIT_SIZE==16
    uint8_t Filler007;              // 007 013 Filler of 1 byte
#else
    uint8_t Filler013[3];           // 007 013 Filler of 3 bytes
#endif
    struct sSUB *pSubscribers;      // 008 016 linked list of subscribers
} PUB;                              // 010 020 Total Size
typedef PUB* pPUB;

typedef struct sQUE {               // Variable Size object
    struct sTCB *pWaitList;         // 000 000 The list of thread waiting on this event
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012 qTYPE_QUE
    uint8_t Status;                 // 007 013 Not used
    uint16_t Size;                  // 008 014 Size of the queue
    unsigned Entries;               // 010 016 Number of entries in the queue
    struct sMSG** pOut;             // 012 020 Index where to read the next message
    struct sMSG** pIn;              // 014 024 Index where to write the next message
    struct sMSG** pMax;             // 016 028 The end of the queue
    struct sMSG* pMessages;         // 018 032 The messages
} QUE;
typedef QUE* pQUE;

__extension__ typedef struct sPIP{  // Variable Size object
    int volatile Entries;           // 000 000 Number of entries in the pipe
    struct sOBJ *pObjList;          // 002 004 Next object in the ObjectList
    char* pName;                    // 004 008 The name of the object
    uint8_t Type;                   // 006 012 qTYPE_PIP
    union {
        uint8_t Status;             // 007 013 Status NOT USED
        uint8_t EtxChar;            // 007 013 ETX character 
    };
    uint8_t Status;                 // 007 013 Status NOT USED
    uint16_t BlockSizeINT;          // 008 014 Size of the datatype or struct in INTEGERS (0 if not a multiply of integer)
    uint16_t BlockSize;             // 010 016 Size of the datatype or struct
    uint16_t MaxBlocks;             // 012 018 The max number of structures it holds - 1
    union {
        void (*pNtfWriter)(struct sPIP*, unsigned); // 014 020 Writing thread that can wait
        struct sTCB *pSignalThread; // 014 020 Thread to signal
    };
    union {
        void (*pNtfReader)(struct sPIP*, unsigned); // 016 024 Reading thread that can wait
        unsigned SignalEvent;       // 016 024 Reading thread that can wait
    };
    uint8_t* pOut;                  // 018 028 Index where to read the next item
    uint8_t* pIn;                   // 020 032 Index where to write the next item
    uint8_t* pMax;                  // 022 036 The address paste the data
    uint8_t Data;                   // 024 040 Data
#if qBIT_SIZE==16
    uint8_t Filler025;              // 025 041 Filler of 1 byte
#else
    uint8_t Filler041[3];           // 025 041 Filler of 3 bytes
#endif
} PIP;                              // 024 034 Total Size
typedef PIP* pPIP;

typedef struct sTMR{                // Fixed Size object
    struct sTCB *pNext;             // 000 000 (Shared TCB) Pointer to next TMR/TCB in the wait/ready list
    struct sOBJ *pObjList;          // 002 004 (Shared OBJ) Next object in the ObjectList
    char* pName;                    // 004 008 (Shared OBJ) The name of the object
    uint8_t Type;                   // 006 012 (Shared TCB/OBJ) The type (qTYPE_TMR)
    uint8_t ListType;               // 007 013 (Shared TCB) 0=NoList, 1=Ready, 2=Timer, 3=Rtcc
    uint8_t TimedOut;               // 008 014 (Shared TCB) 0=NOT-TimedOut 1=TimedOut
    uint8_t Status;                 // 009 015 (Shared TCB) See xTMR_xxxxxxxx
    struct sTCB *Prev;              // 010 016 (Shared TCB) Pointer to previous TMR/TCB in the wait/ready list
    uint32_t WakeUpTime;            // 012 020 (Shared TCB) Alarm Time in cycles
    void (*pFunction)(void*);       // 016 024 The function to call
    int32_t WaitTime;               // 018 028 WaitTime during creation in cycles (Neg for RTCC waits)
    void* pParam;                   // 022 032 Parameter for function called when expired
} TMR;                              // 024 036 Total Size
typedef TMR* pTMR;

typedef struct sMBK {               // Size 2 or more (variable)
    struct sMPL *pMpl;              // 000 000 Pool Address to find the pool
    unsigned Memory;                // 002 004 Memory starts here
} MBK;
typedef MBK* pMBK;

typedef struct sMPL {               // Fixed Size object
    pMBK pFreeBlk;                  // 000 000 Pointer to the next free memory block
    struct sMPL *pNextPool;         // 002 004 Next memory pool
    uint8_t Type;                   // 004 008 The type (qTYPE_MPL)
    uint8_t Filler007;              // 005 009 Filler of 1 byte for 16 and 32 bit
    uint16_t BlockSize;             // 006 010 Size usable memory
    uint16_t Total;                 // 008 012 Total number of blocks currently in use
    #ifdef q32_BIT
        uint16_t Filler014          // 010 014 Filler of 0 byte or 2 bytes
    #endif
} MPL;                              // 010 016 Total Size
typedef MPL* pMPL;

typedef struct sFPL {               // Variable Size object
    unsigned* pFirstBlk;            // 000 000 Pointer to the first memory block
    struct sFPL *pNextPool;         // 002 004 Next memory pool
    uint8_t Type;                   // 004 008 The type (qTYPE_FPL)
    uint8_t Filler007;              // 005 009 Filler of 1 byte
    uint16_t BlockSize;             // 006 010 Size usable memory
    uint16_t Total;                 // 008 012 Total number of blocks in use
    #ifdef q32_BIT
        uint16_t Filler012;         // 010 014 Filler of 2 bytes
    #endif
    unsigned BitTable[];            // 010 016 at least one integer for bit-table 1=free 0=inuse
} FPL;
typedef FPL* pFPL;

typedef struct sSUB {               // Size 6 or 12
    struct sSUB *pNext;             // 000 000 The next subscriber
    void* pObject;                  // 002 004 Pipe or Queue
    void (*pFun)(void*, struct sMSG*);// 004 008 The subscriber function to call
} SUB;                              // 006 012 Total Size
typedef SUB* pSUB;

__extension__ typedef struct sMSG {
    union {
        struct sMSG* pNext;         // 000 000 Pointer to the next message
        struct sFPL* pFixedPool;    // 000 000 Pointer back to the fixed pool
    };
    uint8_t MemPoolType;            // 002 004 DO NOT USE
    uint8_t volatile UseCnt;        // 003 005 UseCount (READ-ONLY)
    uint16_t MsgType;               // 004 006 Type of message (READ-ONLY)
    unsigned Data[];                // 006 008 RO if UseCnt > 1 and RW if UseCnt == 1
} MSG;
typedef MSG* pMSG;

typedef struct sVER {               // Example 2.3-1100
    uint8_t major;                  // Major build number like 2
    uint8_t minor;                  // Minor build number like 3
    uint16_t build;                 // Build number 1100
} VER;
typedef const VER* pVER;

typedef struct sERR {
    uint32_t TrapAddress;           // 000 PC when trap occured
    uint32_t Uptime;                // 004 Uptime in seconds
    pTCB pThread;                   // 008 The current thread
    uint16_t ErrNbr;                // 012 Specifies the last error
    uint8_t StackType;              // 014 Stack Type ThreadStack=0 InterruptStack=1
    uint8_t Zone;                   // 015 0=Thread, 1=Fiber, >1=Interrupt
    unsigned SPLIM;                 // 016 LIMIT of SP
    int InCritical;                 // 020 -1=no critical, 0=first critical etc.
} ERR;                              // 024 total length
typedef ERR* pERR;

typedef struct sBCD_DTM{
    uint8_t year;                   // BCD codification for year, 0x00 to 0x67
    uint8_t rsvd;                   // reserved for future use
    uint8_t mday;                   // BCD codification for day of month 0x01 to 0x31
    uint8_t mon;                    // BCD codification for month 0x01 to 0x31
    uint8_t hour;                   // BCD codification for hours 0x00 to 0x23
    uint8_t wday;                   // BCD codification for day of week 0x00 to 0x06
    uint8_t sec;                    // BCD codification for seconds 0x00 to 0x59
    uint8_t min;                    // BCD codification for minutes 0x00 to 0x59
} BCD_DTM;
typedef BCD_DTM* pBCD_DTM;

__extension__ typedef union {
    struct {
        uint8_t uint8_0;
        uint8_t uint8_1;
    };
    uint16_t uint16;
} uUINT16;
typedef uUINT16* pUINT16;

__extension__ typedef union {
    struct {
        uint8_t uint8_0;
        uint8_t uint8_1;
        uint8_t uint8_2;
        uint8_t uint8_3;
    };
    struct {
        uint16_t uint16_0;
        uint16_t uint16_1;
    };
    uint32_t uint32;
} uUINT32;
typedef uUINT32* pUINT32;

__extension__ typedef union {
    struct {
        uint8_t uint8_0;
        uint8_t uint8_1;
        uint8_t uint8_2;
        uint8_t uint8_3;
        uint8_t uint8_4;
        uint8_t uint8_5;
        uint8_t uint8_6;
        uint8_t uint8_7;
    };
    struct {
        uint16_t uint16_0;
        uint16_t uint16_1;
        uint16_t uint16_2;
        uint16_t uint16_3;
    };
    struct {
        uint32_t uint32_0;
        uint32_t uint32_1;
    };
    uint64_t uint64;                // 64 bits access
} uUINT64;
typedef uUINT64* pUINT64;

extern int qvInCritical;
extern unsigned qvSpawn;

//*************************** All port functions *********************************
//*** Functions that are not inline are in assembler or C modules ****************
//********************************************************************************
static inline __attribute__ ((always_inline)) unsigned qBit0TstSetAtomic(unsigned* Ptr) {
    unsigned res;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	$t8, 0(%1)                      \n" \
	             "   addiu	$t9, $t8, 0                     \n" \
	             "   ori	$t8, $t8, 1                     \n" \
                 "   sc		$t8, 0(%1)                      \n" \
                 "   beq    $t8, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   andi   %0, $t9, 1                      \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                : "=r" (res) : "r" (Ptr)  : "t8", "t9", "memory");                                                                          \
    return res;
}
// The Bit in the qBitMem???Atomic functions set/clr/tst/tgl functions describes the
// bit-number from address 0xbf00000000. To address bit 8 at address 0xbf080010 use
// bit-number 0x08001008. The offset CLR=04 SET=08 and TGL=0C is handled too
static inline __attribute__ ((always_inline)) void qBitMemSetAtomic(unsigned Bit) {
    unsigned b = 1<<(Bit & 0x1f);
    unsigned* p = (unsigned*)(0xbf000008 + (Bit>>8));
    *p = b;
}
static inline __attribute__ ((always_inline)) void qBitMemClrAtomic(unsigned Bit) {
    unsigned b = 1<<(Bit & 0x1f);
    unsigned* p = (unsigned*)(0xbf000004 + (Bit>>8));
    *p = b;
}
static inline __attribute__ ((always_inline)) bool qBitMemTst(unsigned Bit) {
    unsigned* p = (unsigned*)(0xbf000000 + (Bit>>8));
    if (*p & 1<<(Bit & 0x1f)) return true;
    return false;
}
static inline __attribute__ ((always_inline)) void qBitMemTglAtomic(unsigned Bit) {
    unsigned b = 1<<(Bit & 0x1f);
    unsigned* p = (unsigned*)(0xbf00000C + (Bit>>8));
    *p = b;
}
//////////////////////////////////////////////////////////////////////////////////
void qBytClr(void* p, unsigned Size);
void qBytDecAtomic(volatile uint8_t* p);
void qBytIncAtomic(volatile uint8_t* p);
void qBytMov(void* From, void* To, unsigned Size);
void qBytSet(void* p, unsigned Size);
//////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qCrtEnter() {
    IEC0CLR = 2;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
	             "   la 	$t9, qvInCritical               \n" \
                 "1: ll 	$t8, 0($t9)                     \n" \
	             "   addiu	$t8, $t8, 1                     \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                   \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : :  : "t8", "t9", "memory");                                                                          \
}
static inline __attribute__ ((always_inline)) unsigned qCrtEnterFirst() {
    IEC0CLR = 2;
    bool res;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
	             "   la 	$t9, qvInCritical               \n" \
                 "1: ll 	$t8, 0($t9)                     \n" \
	             "   addiu	$t8, $t8, 1                     \n" \
	             "   addiu	$t7, $t8, 0                     \n" \
                 "   sc		$t7, 0($t9)                     \n" \
                 "   beq    $t7, $zero, 1b                  \n" \
                 "   nop                                    \n" \
	             "   addiu	%0, $zero, 1                    \n" \
                 "   movn   %0, $zero, $t8                  \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : "=r" (res) : : "t7", "t8", "t9", "memory");                                                                          \
    return res;
}
static inline __attribute__ ((always_inline)) void qCrtExit() {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
	             "   la 	$t9, qvInCritical               \n" \
                 "1: ll 	$t8, 0($t9)                     \n" \
	             "   addiu	$t8, $t8, -1                    \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                   \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : :  : "t8", "t9", "memory");                                                                          \
    IEC0SET = 2;
}
//////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) uint32_t qDivU3216(uint32_t dividend, uint16_t divisor) {
	return dividend/divisor;
}
static inline __attribute__ ((always_inline)) uint32_t qDivU3216R(uint32_t dividend, uint16_t divisor, uint16_t* Remainder) {
	*Remainder = (uint16_t)(dividend%divisor );
	return dividend/divisor;
}
static inline __attribute__ ((always_inline)) uint64_t qDivU6416(uint64_t dividend, uint16_t divisor) {
	return dividend/divisor;
}
static inline __attribute__ ((always_inline)) uint64_t qDivU6416R(uint64_t dividend, uint16_t divisor, uint16_t* Remainder) {
	*Remainder = (uint16_t)(dividend%divisor );
	return dividend/divisor;
}
static inline __attribute__ ((always_inline)) uint16_t qDivScaling(uint16_t v1, uint16_t v2, uint16_t v3) {
	return (uint16_t)((v1*v2)/v3);
}
static inline __attribute__ ((always_inline)) uint32_t qMulU3216(uint32_t v1, uint16_t v2) {
	return v1*v2;
}
//////////////////////////////////////////////////////////////////////////////////
void qFbrEnqueue0(void (*pFbr)());
void qFbrEnqueue1(void (*pFbr)(void*), void* pPar1);
void qFbrEnqueue2(void (*pFbr)(void*,void*), void* pPar1, void* pPar2);
static inline __attribute__ ((always_inline)) void qFbrSpawn1() {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "   lw     $t9, qvSpawn                    \n" \
                "1:  ll 	$t8, 0($t9)                     \n" \
	             "   ori	$t8, $t8, 1<<3                  \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   la     $t8, IFS0SET                    \n" \
                 "   addi   $t9, $zero, 2                   \n" \
                 "   lw     $t9, 0($t8)                     \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : : : "t8", "t9", "memory");                                                                          \
}
static inline __attribute__ ((always_inline)) void qFbrSpawn2() {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "   lw     $t9, qvSpawn                    \n" \
                "1:  ll 	$t8, 0($t9)                     \n" \
	             "   ori	$t8, $t8, 1<<2                  \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   la     $t8, IFS0SET                    \n" \
                 "   addi   $t9, $zero, 2                   \n" \
                 "   lw     $t9, 0($t8)                     \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : : : "t8", "t9", "memory");                                                                          \
}
static inline __attribute__ ((always_inline)) void qFbrSpawn3() {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "   lw     $t9, qvSpawn                    \n" \
                "1:  ll 	$t8, 0($t9)                     \n" \
	             "   ori	$t8, $t8, 1<<1                  \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   la     $t8, IFS0SET                    \n" \
                 "   addi   $t9, $zero, 2                   \n" \
                 "   lw     $t9, 0($t8)                     \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : : : "t8", "t9", "memory");                                                                          \
}
static inline __attribute__ ((always_inline)) void qFbrSpawn4() {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "   lw     $t9, qvSpawn                    \n" \
                "1:  ll 	$t8, 0($t9)                     \n" \
	             "   ori	$t8, $t8, 1<<0                  \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   la     $t8, IFS0SET                    \n" \
                 "   addi   $t9, $zero, 2                   \n" \
                 "   lw     $t9, 0($t8)                     \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : : : "t8", "t9", "memory");                                                                          \
}
static inline __attribute__ ((always_inline)) void qFbrSpawnRtcc() {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "   lw     $t9, qvSpawn                    \n" \
                "1:  ll 	$t8, 0($t9)                     \n" \
	             "   ori	$t8, $t8, 1<<7                  \n" \
                 "   sc		$t8, 0($t9)                     \n" \
                 "   beq    $t8, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   la     $t8, IFS0SET                    \n" \
                 "   addi   $t9, $zero, 2                   \n" \
                 "   lw     $t9, 0($t8)                     \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 : : : "t8", "t9", "memory");                                                                          \
}
//////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) unsigned qKrnInCritical() {
    return qvInCritical!=-1;            // -1=not in critical
}
static inline __attribute__ ((always_inline)) unsigned qKrnInFiber() {
    unsigned v;
    __asm__ __volatile__("mfc0 %0,$12" : "=r"(v) :  : "memory");
    return (v & 0x0400);	// Interrupt level 1 only
}
unsigned  qKrnInIntStack();
static inline __attribute__ ((always_inline)) unsigned qKrnInIsr() {
    unsigned v;
    __asm__ __volatile__("mfc0 %0,$12" : "=r"(v) :  : "memory");
    return (v & 0xf800)>>10;	// Interrupt level 2 and higher only
}

/////////////////////////////////////////////////////////////////////////////////
void qMemFree(void* p);
static inline __attribute__ ((always_inline)) void qMemFreeFast(void* p) {
    qMemFree(p);
}

////////////////////////////////////////////////////////////////////////////////
void qSemRelease(struct sSEM* Sem);
static inline __attribute__ ((always_inline)) void qSemReleaseFast(struct sSEM* Sem) {
    qSemRelease(Sem);
}
void qSemAcquire(struct sSEM* Sem);
static inline __attribute__ ((always_inline)) void qSemAcquireFast(struct sSEM* Sem) {
    qSemAcquire(Sem);
}
////////////////////////////////////////////////////////////////////////////////
void      qThrDelay(unsigned USec);
int qThrSuspend();
static inline __attribute__ ((always_inline)) void qThrSuspendV4() {
    qThrSuspend();
}
////////////////////////////////////////////////////////////////////////////////
static inline __attribute__ ((always_inline)) void qWrdDecAtomic(volatile int* Ptr) {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	$t9, 0(%0)                      \n" \
	             "   addiu	$t9, $t9, -1                    \n" \
                 "   sc		$t9, 0(%0)                      \n" \
                 "   beq    $t9, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                : : "r" (Ptr) : "t9", "memory");                                                                          \
}
static inline __attribute__ ((always_inline)) bool qWrdDecIfNotZero(volatile int* Ptr) {
    bool res;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	$t9, 0(%1)                      \n" \
	             "   addiu	$t8, $t9, 0                     \n" \
	             "   addiu	$t9, $t9, -1                    \n" \
	             "   movz	$t9, $t8, $t8                   \n" \
                 "   sc		$t9, 0(%1)                      \n" \
                 "   beq    $t9, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 "   addiu	%0, $zero, 1                    \n" \
                 "   movz	%0, $zero, $t8                  \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                : "=r" (res) : "r" (Ptr)  : "t8", "t9", "memory");                                                                          \
    return res;
}
static inline __attribute__ ((always_inline)) void qWrdIncAtomic(volatile int* Ptr) {
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	$t9, 0(%0)                      \n" \
	             "   addiu	$t9, $t9, 1                    \n" \
                 "   sc		$t9, 0(%0)                      \n" \
                 "   beq    $t9, $zero, 1b                  \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                : : "r" (Ptr) : "t9", "memory");                                                                          \
}
void qWrdMov(unsigned* From, unsigned* To, unsigned Size);
void qWrdClr(unsigned* p, unsigned Size);
void qWrdSet(unsigned* p, unsigned Size);
/////////////////////////////////////////////////////////////////////////////////


void yKernelInit();
static inline __attribute__ ((always_inline)) void yKrnSetInterrupt() {
    IFS0SET = 2;
}
unsigned  yMemSetup(unsigned InterruptStackSize);
void      yModeIdle();
void      yModeSleep();
void      yStart();
void      ySwitchThread();
void      qThrDelay(unsigned USec);
unsigned* yThrStkInit(void (*Thread)(void*), void* data, pTCB Tcb, unsigned* Stack, unsigned StackSize);

// Kernel port prototypes
unsigned* yThrStkInit(void (*Thread)(void*), void* data, pTCB Tcb, unsigned* Stack, unsigned StackSize);
unsigned  yMemSetup(unsigned InterruptStackSize);
void      yKernelInit();
void      ySwitchThread();
void      yStart();
pERR      yErrorInit();
void      yErrorReset(unsigned Err);
void      yModeIdle();
void      yModeSleep();
unsigned  yMemSetup(unsigned InterruptStackSize);
void      yModeIdle();
void      yModeSleep();
void      ySwitchThread();
unsigned* yThrStkInit(void (*Thread)(void*), void* data, pTCB Tcb, unsigned* Stack, unsigned StackSize);





// Timer port prototypes
uint8_t   yTimerInit();
void      yTimerSetAlarm(uint32_t Alarm);
uint32_t  yTimerGetTime();
uint32_t  yTimerGetTimeAddUSec(uint32_t USec);
void      yTimerStop();
unsigned  yTimerReshuffleNeeded();
void      yTimerReshuffleEnd();
uint64_t  yGetCycles();



// The RTCC prototypes
uint8_t   yRtccInit() __attribute__((far));
void      yRtccSetDatTim(uint32_t DatTim) __attribute__((far));
uint32_t  yRtccGetDatTimRounded() __attribute__((far));
uint32_t  yRtccGetDatTim() __attribute__((far));
void      yRtccSetAlarm(uint32_t seconds) __attribute__((far));
uint32_t  yRtccGetUptime() __attribute__((far));
//*********************************** End of Port functions ***********************

#define qISR(v)                                             \
    void __attribute__((naked)) _qqqIsr##v() {                                     \
        __asm__ __volatile__(                               \
            ".set noreorder                             \n" \
            ".set noat                                  \n" \
            ".section .vector_" #v ", code              \n" \
            "    j  qHandler_" #v "                     \n" \
            "    rdpgpr $sp, $sp                        \n" \
            "                                           \n" \
            ".section .text                             \n" \
            ".set noreorder                             \n" \
            ".set noat                                  \n" \
            ".global qHandler_" #v "                    \n" \
            "qHandler_" #v ":                           \n" \
            "    lw     $k0, qvIntStack                 \n" \
            "    sltu   $k1, $k0, $sp                   \n" \
            "    movz   $k0, $sp, $k1                   \n" \
            "    addu   $k1, $zero, $sp                 \n" \
            "    addiu  $sp, $k0, -112                  \n" \
            "    sw     $k1, 108($sp)                   \n" \
            "    mfc0   $k1, $14                        \n" \
            "    mfc0   $k0, $13                        \n" \
            "    sw     $k1, 104($sp)                   \n" \
            "    mfc0   $k1, $12                        \n" \
            "    srl    $k0, $k0, 0xa                   \n" \
            "    sw     $k1, 100($sp)                   \n" \
            "    ins    $k1, $k0, 10, 6                 \n" \
            "    mfc0   $k0, $12, 2                     \n" \
            "    ins    $k1, $zero, 1, 4                \n" \
            "    sw     $k0, 96($sp)                    \n" \
            "    mtc0   $k1, $12                        \n" \
            "    sw     $a0, 16($sp)                    \n" \
            "    sw     $a1, 20($sp)                    \n" \
            "    sw     $a2, 24($sp)                    \n" \
            "    sw     $a3, 28($sp)                    \n" \
            "    sw     $t0, 32($sp)                    \n" \
            "    sw     $t1, 36($sp)                    \n" \
            "    sw     $t2, 40($sp)                    \n" \
            "    sw     $t3, 44($sp)                    \n" \
            "    sw     $t4, 48($sp)                    \n" \
            "    sw     $t5, 52($sp)                    \n" \
            "    sw     $t6, 56($sp)                    \n" \
            "    sw     $t7, 60($sp)                    \n" \
            "    sw     $t8, 64($sp)                    \n" \
            "    sw     $t9, 68($sp)                    \n" \
            "    sw     $at, 72($sp)                    \n" \
            "    sw     $ra, 76($sp)                    \n" \
            "    sw     $v0, 80($sp)                    \n" \
            "    sw     $v1, 84($sp)                    \n" \
            "    mfhi   $v0                             \n" \
            "    mflo   $v1                             \n" \
            "    sw     $v0, 88($sp)                    \n" \
            "    jal        qIsr" #v "                  \n" \
            "    sw     $v1, 92($sp)                    \n" \
            "                                           \n" \
            "    lw     $v0, 88($sp)                    \n" \
            "    lw     $v1, 92($sp)                    \n" \
            "    mthi   $v0                             \n" \
            "    mtlo   $v1                             \n" \
            "    lw     $a0, 16($sp)                    \n" \
            "    lw     $a1, 20($sp)                    \n" \
            "    lw     $a2, 24($sp)                    \n" \
            "    lw     $a3, 28($sp)                    \n" \
            "    lw     $t0, 32($sp)                    \n" \
            "    lw     $t1, 36($sp)                    \n" \
            "    lw     $t2, 40($sp)                    \n" \
            "    lw     $t3, 44($sp)                    \n" \
            "    lw     $t4, 48($sp)                    \n" \
            "    lw     $t5, 52($sp)                    \n" \
            "    lw     $t6, 56($sp)                    \n" \
            "    lw     $t7, 60($sp)                    \n" \
            "    lw     $t8, 64($sp)                    \n" \
            "    lw     $t9, 68($sp)                    \n" \
            "    lw     $at, 72($sp)                    \n" \
            "    lw     $ra, 76($sp)                    \n" \
            "    lw     $v0, 80($sp)                    \n" \
            "    lw     $v1, 84($sp)                    \n" \
            "    di                                     \n" \
            "    ehb                                    \n" \
            "    lw     $k0, 96($sp)                    \n" \
            "    mtc0   $k0, $12,2                      \n" \
            "    lw     $k0, 104($sp)                   \n" \
            "    mtc0   $k0, $14                        \n" \
            "    lw     $k0, 100($sp)                   \n" \
            "    lw     $sp, 108($sp)                   \n" \
            "    mtc0   $k0, $12                        \n" \
            "    eret                                   \n" \
            "    nop                                    \n" \
            ".set at                                    \n" \
            ".set reorder                               \n"); \
    } __attribute((__noinline__)) void qIsr##v()

#define qISR_FAST(v)                                        \
    void _qqqIsr##v() {                                     \
        __asm__ __volatile__(                               \
            ".set noreorder                             \n" \
            ".set noat                                  \n" \
            ".section .vector_" #v ", code              \n" \
            "    j  qHandler_" #v "                     \n" \
            "    rdpgpr $sp, $sp                        \n" \
            "                                           \n" \
            ".section .text                             \n" \
            ".set noreorder                             \n" \
            ".set noat                                  \n" \
            ".global qHandler_" #v "                    \n" \
            "qHandler_" #v ":                           \n" \
            "    lw     $k0, qvIntStack                 \n" \
            "    sltu   $k1, $k0, $sp                   \n" \
            "    movz   $k0, $sp, $k1                   \n" \
            "    addu   $k1, $zero, $sp                 \n" \
            "    addiu  $sp, $k0, -44                   \n" \
            "    sw     $k1, 40($sp)                    \n" \
            "    mfc0   $k1, $14                        \n" \
            "    mfc0   $k0, $13                        \n" \
            "    sw     $k1, 36($sp)                    \n" \
            "    mfc0   $k1, $12                        \n" \
            "    srl    $k0, $k0, 0xa                   \n" \
            "    sw     $k1, 32($sp)                    \n" \
            "    ins    $k1, $k0, 10, 6                 \n" \
            "    mfc0   $k0, $12, 2                     \n" \
            "    ins    $k1, $zero, 1, 4                \n" \
            "    sw     $k0, 28($sp)                    \n" \
            "    mtc0   $k1, $12                        \n" \
            "    sw     $v0, 16($sp)                    \n" \
            "    mfhi   $v0                             \n" \
            "    sw     $v0, 20($sp)                    \n" \
            "    mflo   $v0                             \n" \
            "    jal        qIsr" #v "                  \n" \
            "    sw     $v0, 16($sp)                    \n" \
            "                                           \n" \
            "    lw     $v0, 16($sp)                    \n" \
            "    mtlo   $v0                             \n" \
            "    lw     $v0, 20($sp)                    \n" \
            "    mthi   $v0                             \n" \
            "    di                                     \n" \
            "    ehb                                    \n" \
            "    lw     $k0, 28($sp)                    \n" \
            "    mtc0   $k0, $12,2                      \n" \
            "    lw     $k0, 36($sp)                    \n" \
            "    mtc0   $k0, $14                        \n" \
            "    lw     $k0, 32($sp)                    \n" \
            "    lw     $sp, 40($sp)                    \n" \
            "    mtc0   $k0, $12                        \n" \
            "    eret                                   \n" \
            "    nop                                    \n" \
            ".set at                                    \n" \
            ".set reorder                               \n"); \
    } __attribute((__noinline__)) void qIsr##v()



#include "../qKernelCommon.h"

#endif //qKernel_H



