/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  **
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  **
 **  <http://www.quasarsoft.com/license.html>                                       **
 **                                                                                 **
 *************************************************************************************/

#include "qKernel.h"

void yStartThread(void (*task)(void *pd), void *pdata) {
	task(pdata);
	qThrClose(qvCurTcb);
}



unsigned* yThrStkInit(void (*task)(void *pd), void *pdata, pTCB Tcb, unsigned *Tos, unsigned StackSize) {
	unsigned sr;
		
	Tos = Tos + (StackSize>>2);					// stack direction high->low but stack is the low address
	Tcb->StackSize = StackSize;
	__asm__ __volatile__("mfc0 %0,$12" : "=r"(sr) :  : "memory");
	sr |= 0x00000003;							// Set the interrupt enable on and the EXL. The thread will
												// always be started from an ERET. This prevents interrupts 
												// after the status register is restored
	Tos--;
	Tos--;										// two places on the stack for calling the function

	Tcb->a0		= (unsigned)task;
	Tcb->a1		= (unsigned)pdata;
	Tcb->ra		= (unsigned)yStartThread;
	Tcb->SP		= (unsigned)Tos;
	Tcb->SR		= sr;
	Tcb->EPC	= (unsigned)yStartThread;
#ifndef __NO_CHECK__
	Tcb->at		= 0xAAAAAAAA;
	Tcb->v0		= 0x10101010;
	Tcb->v1		= 0x11111111;
	Tcb->a2		= 0xA2A2A2A2;
	Tcb->a3		= 0xA3A3A3A3;
	Tcb->t0		= 0x70707070;
	Tcb->t1		= 0x71717171;
	Tcb->t2		= 0x72727272;
	Tcb->t3		= 0x73737373;
	Tcb->t4		= 0x74747474;
	Tcb->t5		= 0x75757575;
	Tcb->t6		= 0x76767676;
	Tcb->t7		= 0x77777777;
	Tcb->t8		= 0x78787878;
	Tcb->t9		= 0x79797979;
	Tcb->s0		= 0x80808080;
	Tcb->s1		= 0x81818181;
	Tcb->s2		= 0x82828282;
	Tcb->s3		= 0x83838383;
	Tcb->s4		= 0x84848484;
	Tcb->s5		= 0x85858585;
	Tcb->s6		= 0x86868686;
	Tcb->s7		= 0x87878787;
	Tcb->s8		= 0x88888888;
	Tcb->HI		= 0xFFFFFFFF;
	Tcb->LO		= 0xEEEEEEEE;
#endif
    return Tos;     			//Return the stack pointer to the new thread stack
}


// +----------------------------+ 0x0000
// |                            |
// | Static memory              |
// |                            |
// +----------------------------+ 0x0580  C32_heap = yGetMem_Heap()   (_heap)
// |                            |
// | C32 Heap (0x1000)          |         size = yGetMem_HeapSize()   (_min_heap_size)
// |                            |
// +----------------------------+ 0x1580   
// |                            |
// | Interrupt stack  (0x200)   |
// |                            | 0x177c <--- qvIntStack (stack grows down)
// +----------------------------+ 0x1780 <--- Beg
// |                            |
// |                            |
// | Heap                       |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// |                            |
// +----------------------------+ 0x7c00
// |                            |
// | Temporary stack (0x400)    |
// |                            |
// +----------------------------+ 0x8000   last = yGetMem_Last()   (_stack)

//
// This reoutine intializes the memory. It must set the following variables:
// qvMemHeap 		The pointer points to the start of the heap and the function must initialize the 
//					size in the first integer. The heap must be cleared
// qvIntStack		The pointer points to the stack. The stack must be initialized with 0xff
//
// This implementation set qvIntStack as heap + IntStackSize. The rest is heap.
// Because the Sp is initial set to the top of ram therefore we can't clear the heap. The heap is cleared 
// except the highest 1024 bytes. This gives the system time to work. yStart will clear the rest and add to
// the size.
//
unsigned* yGetMem_Last();
unsigned* yGetMem_Heap();
unsigned yGetMem_HeapSize();

unsigned yMemSetup(unsigned InterruptStackSize) {
	unsigned* C32_heap = yGetMem_Heap();
	unsigned* last = yGetMem_Last();
	unsigned C32_heap_size = yGetMem_HeapSize();

	unsigned heapSize;
	qvIntStackLimit = C32_heap + (C32_heap_size>>2);
	qvIntStack = C32_heap + (InterruptStackSize>>2) + (C32_heap_size>>2);
	qvIntStack--;							// 1 smaller so it does not compete with heap
	heapSize = (unsigned)last - (unsigned)qvIntStack;
    heapSize -= 1024;                       // must be corrected in yStart
	qvMemHeap = qvIntStack + 1;             // Heap and stack don't overlap
	qWrdClr(qvMemHeap, heapSize);           // Keep space for now but clear in yStart
	qWrdSet(qvIntStackLimit, InterruptStackSize);
	*qvMemHeap = heapSize;                  // Temporary smaller
	return heapSize;
}
extern void yStart1();
void yStart() {
	unsigned* last = (unsigned*)yGetMem_Last();
	*qvMemHeap += 1024;
	last = last - (1024>>2);
	qWrdClr(last,1024);
	yStart1();
}
void yEnd() {
}
extern unsigned* qvFbrPut;
void qFbrEnqueue0(void (*pFbr)()) {
	unsigned* tmp;
	do {
		tmp = qvFbrPut;
		qvFbrPut = (unsigned*)*tmp++;
	} while (qBit0TstSetAtomic(tmp));
	*tmp++ = ((unsigned)pFbr)+1;
    unsigned* pSpawn = &qvSpawn;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	%0, 0(%1)                       \n" \
	             "   ori	%0, %0, 1<<8                    \n" \
                 "   sc		%0, 0(%1)                       \n" \
                 "   beq    %0, $zero, 1b                   \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 :  "=&r" (tmp) : "r" (pSpawn)  : "memory" );                                                                          \
    IFS0SET = 2;
}
void qFbrEnqueue1(void (*pFbr)(void*), void* pPar1){
	unsigned* tmp;
	do {
		tmp = qvFbrPut;
		qvFbrPut = (unsigned*)*tmp++;
	} while (qBit0TstSetAtomic(tmp));
	*tmp++ = ((unsigned)pFbr)+1;
	*tmp++ = (unsigned)pPar1;
    unsigned* pSpawn = &qvSpawn;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	%0, 0(%1)                       \n" \
	             "   ori	%0, %0, 1<<8                    \n" \
                 "   sc		%0, 0(%1)                       \n" \
                 "   beq    %0, $zero, 1b                   \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 :  "=&r" (tmp) : "r" (pSpawn)  : "memory");                                                                          \
    IFS0SET = 2;
}
void qFbrEnqueue2(void (*pFbr)(void*,void*), void* pPar1, void* pPar2) {
	unsigned* tmp;
	do {
		tmp = qvFbrPut;
		qvFbrPut = (unsigned*)*tmp++;
	} while (qBit0TstSetAtomic(tmp));
	*tmp++ = ((unsigned)pFbr)+1;
	*tmp++ = (unsigned)pPar1;
	*tmp++ = (unsigned)pPar2;
    unsigned* pSpawn = &qvSpawn;
    __asm__ __volatile__(".set noreorder                    \n" \
                 ".set noat                                 \n" \
                 "1: ll 	%0, 0(%1)                       \n" \
	             "   ori	%0, %0, 1<<8                    \n" \
                 "   sc		%0, 0(%1)                       \n" \
                 "   beq    %0, $zero, 1b                   \n" \
                 "   nop                                    \n" \
                 ".set at                                   \n" \
                 ".set reorder                              \n" \
                 :  "=&r" (tmp) : "r" (pSpawn)  : "memory");                                                                          \
    IFS0SET = 2;
}


// this function overrides the normal _weak_ generic handler
extern void SoftReset(void);
void yErrorReset(unsigned Err) {
    unsigned* sp;
    qvErr.ErrNbr = Err;
    qvErr.Uptime = qRtcGetUptime();
    qvErr.pThread = qvCurTcb;
    qvErr.SPLIM = 0;
    __asm__ __volatile__("addiu %0, $sp, 0 \n" : "=r"(sp));
    qvErr.StackType = 0;
    if (sp<qvIntStack) {
        qvErr.StackType = 1;
    }
    qvErr.Zone = qKrnInFiber();
    if (qvErr.Zone!=1) {
        qvErr.Zone = qKrnInIsr();
    }
    qvErr.TrapAddress = 0;
    qvErr.InCritical = qvInCritical;
    RCON = RCON | (1<<6);
    SoftReset();
}

pERR yErrorInit() {
    if (RCON & (1<<6)) {    // if the error was set
        return &qvErr;          // ... return the error field
    }
    RCON = 0;                   // Clear all if not a software reset
    qWrdClr((unsigned*)&qvErr, sizeof(qvErr));
    return NULL;
}

