/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qEvtSignal(pEVT pEvt, unsigned EventFlags) {
    pTCB tcb;
    unsigned thrEvents;
    unsigned clearEvents;
    unsigned ready;
    unsigned ctxSwitch;

#ifndef __NO_CHECK__
    if (!EventFlags ) {
        xErrNtf(qERR_EVT_NO_FLAGS);
        return;
    }
    if (pEvt->Type!= xTYPE_EVT) {
        xErrNtf(qERR_EVT_ID);
        return;
    }
#endif
    if (!qCrtEnterFirst()) {
        if (qKrnInIsr()) {
            qCrtExit();
            qFbrEnqueue2((void (*)(void*,void*))qEvtSignal,pEvt,(void*)EventFlags);
            return;
        }    
    }
    pEvt->CurEvents |= EventFlags;              // Set the events
    ctxSwitch = 0;
    clearEvents = 0;
    
    // Events can make multiple threads ready.
    // For that reason we traverse the whole tree
    tcb = pEvt->pWaitList;                      // Get the start TCB
    while (tcb) {                               // Every thread waiting on this event
        if (tcb->WaitEvtEvents==0) {            // Are we waiting on anything?
            tcb=tcb->pNextWaitObj;              // No so go to the next thread..
            continue;                           // .. that waits on the object
        }
        thrEvents = pEvt->CurEvents & tcb->WaitEvtEvents;    // Events important for this thread
        if (tcb->Status & xTCB_WAIT_TYPE_OR)    // OR case
            ready = (thrEvents!=0);
        else                                    // AND case
            ready = (thrEvents==tcb->WaitEvtEvents);
        if (ready) {
            tcb->WaitEvtEvents = thrEvents;     // The events that made the thread ready
            if (tcb->Status & xTCB_WAIT_AND_CLEAR)
                clearEvents |= thrEvents;       // Masks of bits to clear
            xObjRemove(tcb);                    // remove from the object list
            if (xRemWaitInsRdyList(tcb))        // Remove from wait Insert in ready
                ctxSwitch = 1;                  // On the top of the list so we have to
        }                                       // do a context switch
        tcb=tcb->pNextWaitObj;                  // next thread that waits on the object
    }
    pEvt->CurEvents &= ~clearEvents;            // Clear all the events that are set
    if (ctxSwitch) {
        if (qKrnInIsr()) {
            yKrnSetInterrupt();                 // Use the kernel interrupt to do that because
        }
        else {
            ySwitchThread();                    // Switch thread and EXIT_CRITICAL
            return;                             // only if readied thread > prio
        }
    }
    qCrtExit();
}
