/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"
static uint8_t* CalcTracking(unsigned* pAdr, unsigned Bit, uint8_t* pBitValue);


void qKrnTrackingIdle(unsigned *pIdleAdr, unsigned IdleBitNbr) {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
	if (IdleBitNbr >= sizeof(int)*8) {
		xErrNtf(qERR_KRN_IDLE_BIT_NBR);
		return;
	}
#endif
    qvTrackingIdleAdr  = CalcTracking(pIdleAdr,  IdleBitNbr,  &qvTrackingIdleBit);
}

void qKrnTrackingRun(unsigned *pRunAdr, unsigned RunBitNbr) {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
	if (RunBitNbr >= sizeof(int)*8) {
		xErrNtf(qERR_KRN_RUN_BIT_NBR);
		return;
	}
#endif
    qvTrackingRunAdr   = CalcTracking(pRunAdr,   RunBitNbr,   &qvTrackingRunBit);
}
void qKrnTrackingSleep(unsigned *pSleepAdr, unsigned SleepBitNbr) {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
	if (SleepBitNbr >= sizeof(int)*8) {
		xErrNtf(qERR_KRN_SLEEP_BIT_NBR);
		return;
	}
#endif
    qvTrackingSleepAdr = CalcTracking(pSleepAdr, SleepBitNbr, &qvTrackingSleepBit);
}


static uint8_t* CalcTracking(unsigned* pAdr, unsigned Bit, uint8_t* pBitValue) {
    uint8_t* p = (uint8_t*)pAdr;        // We use a byte pointer
    uint8_t bitValue;
    if (p==0) {                         // Disable thread tracking for this thread
        bitValue = 0;                   // IOR will do nothing with a value of 0
        p = pBitValue;
    }
    else {
        while (Bit>7) {                 // Change integer pointer to char pointer
            Bit-=8;
            p++;
        }
        bitValue = 1<<Bit;
    }
    *pBitValue = bitValue;
    return p;
}

