/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"



static pMSG xMsgAllocFast(pMPL pMpl, uint16_t MaxSize, uint16_t MsgType) {
    pMSG p;
    p = qMemAllocFromPoolClr(pMpl);
    if (p != 0) {
        uint8_t sizeDiff = (pMpl->BlockSize-MaxSize)<<4;
        p->UseCnt = 1;
        p->MemPoolType = sizeDiff + xTYPE_MPL_MSG;
        p->MsgType = MsgType;
        p->pNext = qvMsgList;       // Put the new message at the beginning at ...
        qvMsgList = p;              // ...message list.
    }
    return p;
}

pMSG qMsgAlloc(uint16_t MaxSize, uint16_t MsgType) {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_MSG_ISR);
        return NULL;
    }
    if (MaxSize<sizeof(struct sMSG) || MaxSize>32000) {
        xErrNtf(qERR_MSG_SIZE);
        return NULL;
    }
#endif
    pMPL mpl = qMemPool(MaxSize);    // pool + size of message
    if (mpl==NULL) return NULL;
    return xMsgAllocFast(mpl, MaxSize, MsgType);
}

pMSG qMsgAllocFast(pMPL pMpl, uint16_t MaxSize, uint16_t MsgType) {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_MSG_ISR);
        return NULL;
    }
    if (pMpl==NULL || pMpl->Type!=xTYPE_MPL) {
        xErrNtf(qERR_MEM_ID);
        return 0;
    }
    if (MaxSize > pMpl->BlockSize || MaxSize < (pMpl->BlockSize-15)) {
        xErrNtf(qERR_MSG_SIZE);
        return 0;
    }
#endif
    return xMsgAllocFast(pMpl, MaxSize, MsgType);
}

