/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qMsgPublish(pPUB pPub, pMSG pMsg) {
#ifndef __NO_CHECK__
    if (pMsg==NULL) {
        xErrNtf(qERR_MSG_ID);
        return;
    }
    uint8_t poolType = pMsg->MemPoolType & 0xf; // Keep only the least significant nibble
    if (poolType!=xTYPE_MPL_MSG && poolType!=xTYPE_FPL_MSG) {
        xErrNtf(qERR_MSG_ID);
        return;
    }
    if (pPub==NULL || pPub->Type!=xTYPE_PUB) {
        xErrNtf(qERR_PUB_ID);
        return;
    }
#endif

    if (!qCrtEnterFirst()) {
        if (qKrnInIsr()) {
            qCrtExit();
            qFbrEnqueue2((void (*)(void*,void*))qMsgPublish, pPub, pMsg);
            return;
        }    
    }
    pSUB s = pPub->pSubscribers;                // next (first) subscriber
    while (s) {
        if (s->pObject==0) {                    // if 0 then function (no Pipe/queue) so
            qBytIncAtomic(&pMsg->UseCnt);       // we have to increase the use-count
        }                                       // Pipes and queues do it on send and write
        (s->pFun)(s->pObject, pMsg);            // call the function with object (pipe/Queue) and message
        s = s->pNext;                           // next subscriber
    }    
    qCrtExit();
}



