/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


unsigned qMsgWrite(pPIP pPip, pMSG pMsg) {
#ifndef __NO_CHECK__
    if (pPip->Type != xTYPE_PIP) {
        xErrNtf(qERR_PIP_ID);
        return 0;
    }
    if (pMsg==NULL) {
        xErrNtf(qERR_MSG_ID);
        return 0;
    }
    uint8_t poolType = pMsg->MemPoolType & 0xf; // Keep only the least significant nibble
    if (poolType!=xTYPE_MPL_MSG && poolType!=xTYPE_FPL_MSG) {
        xErrNtf(qERR_MSG_ID);
        return 0;
    }
    if (pPip->BlockSizeINT!=1) {
        xErrNtf(qERR_MSG_PIP_SIZE);
        return 0;
    }

#endif
    unsigned result;
    qBytIncAtomic(&pMsg->UseCnt);
    // Ignore the warning "Taking the address of 'pMsg' may require an extended pointer for this device
    // Q-Kernel never allocates the stack in memory above 0x8000
    result = qPipWrite(pPip, &pMsg, 1);
    if (result==0) {                            // if not written it can not be used by 
        qBytDecAtomic(&pMsg->UseCnt);           // another thread so undo incrementing use-count
    }    
    return result;
}
    
