/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qMtxLock(pMTX pMtx) {
    pTCB owner;
    uint8_t prio;

#ifndef __NO_CHECK__
    if (xTestWait(pMtx, xTYPE_MTX, qERR_MTX_ID, 1000)) { // 1000 so it test waiting conditions
        return;
    }
#endif
    
    if (pMtx->Owner==qvCurTcb) return;          // We already own the mutex so we are done
    qCrtEnter();
    if (pMtx->Owner==0 || pMtx->Owner==qvCurTcb) { // Is the mutex available or we already own it then lock it
        pMtx->Owner = qvCurTcb;                 // We are the owner so store us
        pMtx->OrgPrio = qvCurTcb->Prio;         // and the original priority
        qCrtExit();
        return;
    }
//
// Here if another thread has locked the mutex
//
    prio = 0;                                   // Expect that we don't have to increase the ...
    owner = pMtx->Owner;                        // ... priority of the owner
    qvCurTcb->pObject = (pOBJ)pMtx;             // We want to wait on this object
    xRemRdyList(qvCurTcb);                     // Remove us from the ReadyList
    if (xObjInsert(qvCurTcb) && qvCurTcb->Prio > owner->Prio) { 
                                                // Are we inserted at the highest priority
                                                // and is our priority higher then the owner
        owner->Prio = qvCurTcb->Prio;           // give it our priority
        if (owner->ListType==xLIST_READY) {     // is the owner NOT in a waitlist
            xRemRdyList(owner);                // Remove from the RdyList and add the thread ...
            xInsRdyList(owner);                 // ... again with the new priority
        }
    }
    if (qvCurTcb!=qvRdyList) {
        ySwitchThread();                        // Switch Thread and EXIT_CRITICAL
        qCrtEnter();
    }
    pMtx->Owner = qvCurTcb;                     // We are the owner so store us
    pMtx->OrgPrio = qvCurTcb->Prio;             // and the original priority
    qCrtExit();
}

