/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qMtxUnlock(pMTX pMtx) {
    pTCB tcb;
    
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_MTX_ISR);
        return;
    }
    if (pMtx==NULL || pMtx->Type!= xTYPE_MTX) {
        xErrNtf(qERR_MTX_ID);
        return;
    }
    if (qKrnInFiber()) {
        xErrNtf(qERR_MTX_FBR);
        return;
    }
    if (!xKrnIsOsStarted()) {
        xErrNtf(qERR_MTX_NO_START);
        return;
    }
#endif

    qCrtEnter();
#ifndef __NO_CHECK__
    if (qvCurTcb != pMtx->Owner) {               // See if the current thread owns the mutex
        xErrNtf(qERR_MTX_OWNER);
        return;
    }
#endif
    
    pMtx->Owner = 0;                            // we don't own the mutex anymore
    if (qvCurTcb->Prio != pMtx->OrgPrio) {      // did somebody increased our priority?
        xRemRdyList(qvCurTcb);                 // Yes, remove us from the current prio
        qvCurTcb->Prio = pMtx->OrgPrio;         // set the new priority
        xInsRdyList(qvCurTcb);                  // And add us again with the new prio
    }

    tcb = pMtx->pWaitList;                      // List of thread that we wait for in prio order
    if (tcb) {                                  // Are there threads waiting? (thread is readied)
        xObjRemove(tcb);                        // remove from the object list
        if (xRemWaitInsRdyList(tcb)) {          // Remove WaitList Insert ReadyList
            ySwitchThread();                    // Switch thread and EXIT_CRITICAL
            return;                             // only if readied thread > prio
        }
    }
    qCrtExit();
}


