/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"
                             

pQUE qQueCreate(char* pName, uint16_t QueueSize){
    pQUE p;
    unsigned size;

#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_QUE_ISR);
        return NULL;
    }
    if (xErrNtf(xObjCheck(pName, qERR_QUE_ID, xTYPE_QUE))) {
        return NULL;
    }
    if (QueueSize==0) {
        xErrNtf(qERR_QUE_SIZE);
        return NULL;
    }
#endif
                                    
    qCrtEnter();
    size = sizeof(struct sQUE);    
    size +=(QueueSize-1)*sizeof(void*);         // structure contains one entry
    p = xObjAlloc(xTYPE_QUE,size);
    if (!p) {
        xErrNtf(qERR_QUE_MEMORY);
        return NULL;
    }

    p->pName = pName;                           // Store the name
    p->Type = xTYPE_QUE;                        // allocate the object by writing type
    p->pMax = &p->pMessages + QueueSize;        // pMax positioned after the last message
    p->Size = QueueSize;
    p->pIn = &p->pMessages;                     //
    p->pOut = &p->pMessages;                    //
    qCrtExit();
    xObjRemOpenList(pName, xTYPE_OPN_QUE );     // Make waiting Open threads runnable
    return p;
}


void xQueReadyWaiting(pQUE p) {
//
// Ready senders or receivers (MUST BE IN CRITICAL SECTION)
//
    pTCB tcb = p->pWaitList;                // List of waiting threads (prio order)
    if (tcb) {
        xObjRemove(tcb);
        if (xRemWaitInsRdyList(tcb)) {      // Remove WaitList Insert ReadyList
            ySwitchThread();                // Switch Thread and EXIT_CRITICAL
                                            // only if readied thread > prio
            return;
        }
    }
    qCrtExit();
    return;
}    
