/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

void qThrEvtSignal(pTCB pTcb, unsigned EventFlags) {
    unsigned thrEvents;
    unsigned clearEvents;
    unsigned ready;
    unsigned ctxSwitch;


#ifndef __NO_CHECK__
    if (pTcb==0 || pTcb->Type!=xTYPE_TCB) {
        xErrNtf(qERR_THR_ID);
        return;
    }
    if (!EventFlags ) {
        xErrNtf(qERR_THR_NO_FLAGS);
        return;
    }
#endif
    if (!qCrtEnterFirst()) {
        if (qKrnInIsr()) {
            qCrtExit();
            qFbrEnqueue2((void (*)(void*,void*))qThrEvtSignal, pTcb, (void*)EventFlags);
            return;
        }
    }

    pTcb->CurThrEvents |= EventFlags;           // Set the events
    clearEvents = 0;
    ctxSwitch = 0;
    
    if (pTcb->pObject == xTHR_EVT_WAITING) {    // Are we waiting on anything?
        thrEvents = pTcb->CurThrEvents & pTcb->WaitThrEvents; // Events important for this thread
        if (pTcb->Status & xTCB_WAIT_TYPE_OR)   // OR or OR_CLEAR case
            ready = (thrEvents!=0);
        else                                    // AND or AND_CLEAR case
            ready = (thrEvents==pTcb->WaitThrEvents);
        if (ready) {
            pTcb->WaitThrEvents = thrEvents;    // Save the events that made the thread ready
            if (pTcb->Status & xTCB_WAIT_AND_CLEAR) {
                clearEvents |= thrEvents;       // Mask of bits to clear
            }
            pTcb->pObject = 0;                  // Indicate the Thread Event hwas triggered
            if (pTcb->ListType!=xLIST_READY) {  // Is the thread waiting?
                if (xRemWaitInsRdyList(pTcb)) { // Remove from wait Insert in ready
                    ctxSwitch = 1;              // If priority higher indicate we have
                }                               // to reschedule
            }
            pTcb->CurThrEvents &= ~clearEvents;     // Clear all the events that are set
        }
    }
    
    if (ctxSwitch==1) {                         // Do we need to reschedule
        if (qKrnInIsr()) {
            yKrnSetInterrupt();                 // Use the kernel imterrupt to do that because
        }                                       // we cannot use switchThread in an interrupt.
        else {
            ySwitchThread();                    // Switch thread and EXIT_CRITICAL
            return;                             // only if readied thread > prio
        }
    }
    qCrtExit();
}
