/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

unsigned qThrEvtWait(unsigned EventFlags, WAIT_TYPE WaitType) {
    unsigned events;
    unsigned ready;
    
#ifndef __NO_CHECK__
    if (xTestWait(0, xTYPE_TCB, qERR_THR_ID, 1000)) // 1000 so it test waiting conditions
        return 0;
    if (!EventFlags) {
        xErrNtf(qERR_THR_NO_FLAGS);
        return 0;
    }
    if (WaitType < WAIT_TYPE_ALL || WaitType > WAIT_TYPE_ANY_AND_CLEAR) {
        xErrNtf(qERR_THR_WAIT_TYPE);
        return 0;
    }
#endif
    qCrtEnter();
    ready = 0;
    events = qvCurTcb->CurThrEvents&EventFlags; // events we are interested in
    if (WaitType&1) {                           // ANY/OR case in Bit 0
        if (events) ready=1;
    }
    else {                                      // ALL/AND case
        if (events==EventFlags) ready=1;
    }
    if (ready) {
        if (WaitType&2) {                       // Clear in Bit 1
            qvCurTcb->CurThrEvents &= ~events;  // clear the events that were set
        }
        qCrtExit();
        return events;                          // return success
    }

    qvCurTcb->WaitThrEvents = EventFlags;       // Waiting on these events
    qvCurTcb->Status &= ~xTCB_WAIT_TYPE_OR;
    qvCurTcb->Status &= ~xTCB_WAIT_AND_CLEAR;
    qvCurTcb->pObject = xTHR_EVT_WAITING;
    if (WaitType&1) qvCurTcb->Status |= xTCB_WAIT_TYPE_OR;
    if (WaitType&2) qvCurTcb->Status |= xTCB_WAIT_AND_CLEAR;
    
    xRemRdyList(qvCurTcb);                      // Remove us from the ReadyList in the waitlist
    ySwitchThread();                            // Switch Thread and EXIT_CRITICAL
    return qvCurTcb->WaitThrEvents;             // contains the events that readied thread
}
