/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qThrResume(pTCB pTcb, int reason) {
#ifndef __NO_CHECK__
    if (pTcb==0 || pTcb->Type!=xTYPE_TCB) {
        xErrNtf(qERR_THR_ID);
        return;
    }
    if (pTcb->pObject && (pTcb->pObject != xTHR_EVT_WAITING)) {
        xErrNtf(qERR_THR_WAIT_OBJECT);
        return;
    }
    if (reason==-1) {
        xErrNtf(qERR_THR_REASON);
        return;
    }
#endif
    if (!qCrtEnterFirst()) {
        if (qKrnInIsr()) {
            qCrtExit();
            qFbrEnqueue2((void (*)(void*, void*))qThrResume, pTcb, (void*)reason);
            return;
        }
    }    
    if (pTcb->ListType>xLIST_READY) {           // Is it in one of the wait lists
        if (xRemWaitInsRdyList(pTcb)) {         // Remove from wait Insert in ready
            pTcb->WakeUpTime = (uint32_t)reason; // Store for output to the sleep
            if (qKrnInIsr()) {
                yKrnSetInterrupt();             // Use the kernel imterrupt to do that because
            }                                   // we cannot use switchThread in an interrupt.
            else {
                ySwitchThread();                // Switch thread and EXIT_CRITICAL
                return;                         // only if readied thread > prio
            }
        }
        qCrtExit();
    }
    else if (pTcb->ListType==xLIST_NONE) {
        if (xInsRdyList(pTcb)) {                // Insert in ready
            pTcb->WakeUpTime = (uint32_t)reason; // Store for output to the sleep
            if (qKrnInIsr()) {
                yKrnSetInterrupt();             // Use the kernel imterrupt to do that because
            }                                   // we cannot use switchThread in an interrupt.
            else {
                ySwitchThread();                // Switch thread and EXIT_CRITICAL
                return;                         // only if readied thread > prio
            }
        }
        qCrtExit();
    }
    else {                                      // This is the case that the thread is in the ready state
        qCrtExit();
    }
}


