/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


pTMR qTmrCreate(char* pName, void (*pFunction)(void*), void* pParam, TIMER_TYPE TimerType, int32_t WaitTime) {
    pTMR p;
    bool periodic = false;
    bool manualStart = false;
    if (TimerType==TIMER_TYPE_PERIODIC || TimerType==TIMER_TYPE_PERIODIC_MANUAL_START) {
        periodic = true;
    }
    if (TimerType==TIMER_TYPE_ONE_SHOT_MANUAL_START || TimerType==TIMER_TYPE_PERIODIC_MANUAL_START) {
        manualStart = true;
    }
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_TMR_ISR);
        return NULL;
    }
    if (xErrNtf(xObjCheck(pName, qERR_TMR_ID, xTYPE_TMR))) {
        return NULL;
    }
    if (pFunction==0) {
        xErrNtf(qERR_TMR_FUNCTION);
        return NULL;
    }
    if (WaitTime>=0 && WaitTime<20) {       // less than 20uSec doesn't make sense
        xErrNtf(qERR_TMR_TIME);
        return NULL;
    }
    if (!xKrnIsOsStarted() && !manualStart) {
        xErrNtf(qERR_TMR_NO_START);
        return NULL;
    }

#endif

    p = xObjAlloc(xTYPE_TMR, sizeof(struct sTMR));
    if (!p) {
        xErrNtf(qERR_TMR_MEMORY);
        return NULL;
    }
    qCrtEnter();
    p->Type = xTYPE_TMR;
    p->pName = pName;
    p->pFunction = pFunction;
    p->pParam = pParam;
    if (WaitTime<0) {
        p->WaitTime = WaitTime;
    }
    else {
        p->WaitTime = xUSec2Cycles(WaitTime);
    }
    if (periodic) {
        p->Status = xTMR_PERIODIC;
    }
    if (!manualStart) {
        qTmrStart(p);
    }
    qCrtExit();
    xObjRemOpenList(pName, xTYPE_OPN_TMR); // Make waiting Open threads runnable
    return p;
}

