/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void qTmrStart(pTMR pTmr) {
#ifndef __NO_CHECK__
    if (pTmr==NULL || pTmr->Type!= xTYPE_TMR) {
        xErrNtf(qERR_TMR_ID);
        return;
    }
    if (!xKrnIsOsStarted()) {
        xErrNtf(qERR_TMR_NO_START);
        return;
    }
#endif    
        
    if (!qCrtEnterFirst()) {
        if (qKrnInIsr()) {
            qCrtExit();
            qFbrEnqueue1((void(*)(void*))qTmrStart, pTmr);    
            return;
        }
    }
    pTmr->Status &= ~(xTMR_STOP);               // Remove the stop so we always continue
    if (pTmr->ListType!=xLIST_NONE) {           // Are we already running?
        qCrtExit();                             // Yes we are done
        return;
    }    
    if (pTmr->WaitTime<0) {
        pTmr->WakeUpTime = yRtccGetDatTimRounded() + (-pTmr->WaitTime);
        if (xInsRtccList((pTCB)pTmr))           // Insert in the RTCC list and set alarm if first in list
            yRtccSetAlarm(pTmr->WakeUpTime);
    }
    else {
        pTmr->WakeUpTime = yTimerGetTime() + pTmr->WaitTime;
        if (xInsTimerList((pTCB)pTmr)) {        // Insert in the Timer list and set alarm if first in list
            yTimerSetAlarm(pTmr->WakeUpTime);
        }    
    }
    qCrtExit();
}
