/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


void xFbrTimerExpire() {
    pTCB p;
    unsigned newAlarmRequired;
    
/*
    This fiber is called from the Timer Interrupt and the TimerSetAlarm function. If the source is the 
    Timer Interrupt the timer is not set so we have to set it or when the timer list is empty we
    have to stop the timer. If the source is the TimerSetAlarm function it just has to verify if the
    time to set it was to close so the interrupt did not fire. In that case we have to set the alarm only when 
    we execute an expire (going into the while loop)
    We run under ** level 1 so we don't have to enter critical sections. This routine will NOT 
    be called if thread Switching is disabled, so we don't have to use critical sections. The caller 
    will do the context-switch if required.
*/
    newAlarmRequired=xKrnIsTimerInterruptOn();
    if (newAlarmRequired != 0) {
                                                // is the source a timer interrupt then new alarm required
        xKrnClrTimerInterruptOn();              // and clear the flag
    }    
/*
    Walk true the wait list and see if a thread is timed-out. We do this until we reach wakeup times
    that are bigger then the Timer. If that's the case we are done.
*/
    while (((p=qvTimerList)!=0) && (p->WakeUpTime<=yTimerGetTime())) {
        newAlarmRequired = 1;
        if (p->Type==xTYPE_TCB) {
            p->TimedOut = 1;                    // Indicate a time out
            if (p->pObject) {                   // Do we time out from a object (not a Sleep)
                xObjRemove(p);                  // Yes then remove us from the object
            }
            xRemWaitInsRdyList(p);              // Remove from wait Insert in ready
        }
        else {
            pTMR tmr = (pTMR)p;                 // change the type to timer
            (tmr->pFunction)(tmr->pParam);      // call the timer
            xRemWaitList((pTCB)tmr);
            if (tmr->Status&xTMR_DELETE) {
                xObjFree((pOBJ)tmr);            // Remove from the timer pool
            }                                
            else {
                if (tmr->Status&xTMR_PERIODIC && (tmr->Status&xTMR_STOP)==0) {
                    tmr->WakeUpTime += tmr->WaitTime;
                    xInsTimerList((pTCB)tmr);   // Insert (Remove was done above)
                }
            }
        }        
    }
    if (qvTimerList) {
        if (newAlarmRequired) {
            yTimerSetAlarm(qvTimerList->WakeUpTime);
        }    
    }    
    else {
        yTimerStop();                           // If timing then we can stop the timer
    }
/*
    We resuffel all WakeUp values in the Threads and timers in the TimerList. The reshuffle function disables
    the timer interrupt and subtracts 0x80000000 from the alarm and time. 
*/
    if (yTimerReshuffleNeeded()) {              // Resuffel Needed (Disable interrupts if yes)
        p = qvTimerList;                        // Timer list for Timers and Thread
        while (p) {                             // All in the wait list because it does not contain hibernated once
            p->WakeUpTime -= 0x80000000;        // Subtract 0x8000.0000 so we can always wait 0x7fff.ffff periods...
            p = p->pNext;                       // ... without overflow and start over again
        }
        yTimerReshuffleEnd();                   // Will enable interrupts again
    }
}    
