/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"


pOBJ xObjOpen(char* pName, unsigned ErrId, uint8_t Type) {
    pOBJ p;
    uint8_t OpenType;

#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(ErrId+2);                       // XXX_ISR
        return NULL;
    }
    if (!xKrnIsOsStarted()) {
        xErrNtf(ErrId+1);                       // XXX_NO_START
        return NULL;
    }
    if (qKrnInFiber()) {
        xErrNtf(ErrId+3);                       // XXX_FBR
        return NULL;
    }
    if (pName==0) {
        xErrNtf(ErrId+8);                       // XXX_N0_NAME
        return NULL;
    }
    if (qKrnInCritical()) {
        xErrNtf(ErrId+9);                       // XXX_CRITICAL
        return NULL;
    }    
#endif

    qCrtEnter();
    if (Type == xTYPE_TCB)
        p = (pOBJ)xGetTcb(pName);
    else
        p = xObjSearch(pName, Type);            // get the object
    if (p) {
        qCrtExit();
        return p;
    }    

    OpenType = Type + 0x80;                     // change from normal type to _OPN_ type
    p = xObjSearch(pName, OpenType);            // something else waiting 
    if (!p) {                                   // Nothing currently waiting
        p = xObjAlloc(OpenType,0);
        if (!p) {
            xErrNtf(ErrId+6);
            return NULL;
        }
        p->pName = pName;
        p->Type = OpenType;
    }

    xThrSuspendOnObject(p);                     // We wait on a object EXIT_CRITICAL

    qCrtEnter();
    if (Type == xTYPE_TCB)
        p = (pOBJ)xGetTcb(pName);
    else
        p = xObjSearch(pName, Type);            // get the object
    qCrtExit();
    return p;
}

pEVT qEvtOpen(char* pName) {
    return (pEVT)xObjOpen(pName, qERR_EVT_ID, xTYPE_EVT);
}

pMTX qMtxOpen(char* pName) {
    return (pMTX)xObjOpen(pName, qERR_MTX_ID, xTYPE_MTX);
}

pPIP qPipOpen(char* pName) {
    return (pPIP)xObjOpen(pName, qERR_PIP_ID, xTYPE_PIP);
}

pQUE qQueOpen(char* pName) {
    return (pQUE)xObjOpen(pName, qERR_QUE_ID, xTYPE_QUE);
}

pSEM qSemOpen(char* pName) {
    unsigned p =(unsigned)xObjOpen(pName, qERR_SEM_ID, xTYPE_SEM);
    p = p + 4*sizeof(int);
    return (pSEM)p;
}

pTCB qThrOpen(char* pName) {
    return (pTCB)xObjOpen(pName, qERR_THR_ID, xTYPE_TCB);
}

pTMR qTmrOpen(char* pName) {
    return (pTMR)xObjOpen(pName, qERR_TMR_ID, xTYPE_TMR);
}

pPUB qPubOpen(char* pName) {
    return (pPUB)xObjOpen(pName, qERR_PUB_ID, xTYPE_PUB);
}

