/*************************************************************************************
 **  Q-KernelFree  Copyright (C) 2013 QuasarSoft Ltd.                               **
 **                                                                                 **
 **  This file is part of Q-KernelFree distribution.                                **
 **  Q-KernelFree is free software:  you can redistribute it and/or modify it under **
 **  the terms of the  GNU  General  Public  License  (version 3) as published  by  **
 **  the Free Software  Foundation ***** AND MODIFIED BY ***** the QuasarSoft Ltd.  **
 **  exception.                                                                     **
 **                                                                                 **
 **    *********************** The QuasarSoft Ltd. EXCEPTION ******************     **
 **    * You may not exercise any of the rights granted to You in any manner  *     **
 **    * that is primarily intended for or directed toward commercial         *     **
 **    * advantage or private monetary compensation. The exchange of the      *     **
 **    * Program for other copyrighted works by means of digital file-sharing *     **
 **    * or otherwise shall not be considered to be intended for or directed  *     **
 **    * toward commercial advantage or private monetary compensation,        *     **
 **    * provided there is no payment of any monetary compensation in         *     **
 **    * connection with the exchange of copyrighted works.                   *     **
 **    ************************************************************************     **
 **                                                                                 **
 **  Q-KernelFree is distributed in the hope that it will be useful,  but  WITHOUT  **
 **  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  ** 
 **  FOR  A  PARTICULAR  PURPOSE.  See the full license text at the following link  ** 
 **  <http://www.quasarsoft.com/license.html>                                       ** 
 **                                                                                 **
 *************************************************************************************/  

#include "qKernel.h"

static void DummyUSecTimer(void* dummy) {
}



void qKrnUSecOn() {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
#endif
    qvCycles = 0;
    if (!xKrnIsUSecOn()) {
        xKrnSetUSecOn();
        qTmrCreate("DummyUSecTimer", DummyUSecTimer, NULL, TIMER_TYPE_PERIODIC, 0x1000000);
    }
}



void qKrnUSecOff() {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_KRN_ISR);
        return;
    }
#endif
    if (xKrnIsUSecOn()) {
        qTmrClose(qTmrOpenNB("DummyUSecTimer"));
        xKrnClrUSecOn();
    }
}


uint64_t qTimCycles() {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_TMR_ISR);
        return 0;
    }
    if (!xKrnIsUSecOn()) {
        xErrNtf(qERR_TMR_NO_USEC);
        return 0;
    }
#endif
    return yGetCycles();
}


uint64_t qTimUSec() {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_TMR_ISR);
        return 0;
    }
    if (!xKrnIsUSecOn()) {
        xErrNtf(qERR_TMR_NO_USEC);
        return 0;
    }
#endif
    uint64_t cycles = yGetCycles();
    if (qvClockFrequencyMhzPower2) {
        return cycles>>qvClockFrequencyMhzPower2;
    }
    return qDivU6416(cycles<<8,qvClockFrequencyMul256Div1000000);
}

uint64_t qTimMSec() {
#ifndef __NO_CHECK__
    if (qKrnInIsr()) {
        xErrNtf(qERR_TMR_ISR);
        return 0;
    }
    if (!xKrnIsUSecOn()) {
        xErrNtf(qERR_TMR_NO_USEC);
        return 0;
    }
#endif
    uint64_t cycles = yGetCycles();
    cycles = qDivU6416(cycles<<8,qvClockFrequencyMul256Div1000000);
    return qDivU6416(cycles,1000);
}

